import typing
from contextlib import suppress

import numpy as np
import pandas as pd
from napari import Viewer
from napari.layers import Layer
from napari.utils.notifications import show_info
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QDialog,
    QGridLayout,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPlainTextEdit,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from PartSeg import plugins
from PartSeg._roi_analysis.profile_export import ExportDialog, ImportDialog, ProfileDictViewer
from PartSeg.common_backend.base_settings import IO_SAVE_DIRECTORY, BaseSettings
from PartSeg.common_backend.except_hook import show_warning
from PartSeg.common_gui.algorithms_description import (
    AlgorithmChooseBase,
    FormWidget,
    InteractiveAlgorithmSettingsWidget,
)
from PartSeg.common_gui.custom_load_dialog import PLoadDialog
from PartSeg.common_gui.custom_save_dialog import PSaveDialog
from PartSeg.common_gui.searchable_combo_box import SearchComboBox
from PartSeg.common_gui.searchable_list_widget import SearchableListWidget
from PartSeg.common_gui.universal_gui_part import TextShow
from PartSeg.plugins import register as register_plugins
from PartSeg.plugins.napari_widgets._settings import get_settings
from PartSeg.plugins.napari_widgets.utils import NapariFormWidgetWithMask, generate_image
from PartSegCore import UNIT_SCALE, Units
from PartSegCore.algorithm_describe_base import AlgorithmSelection, ROIExtractionProfile
from PartSegCore.analysis.algorithm_description import AnalysisAlgorithmSelection
from PartSegCore.analysis.load_functions import LoadProfileFromJSON
from PartSegCore.analysis.save_functions import SaveProfilesToJSON
from PartSegCore.mask.algorithm_description import MaskAlgorithmSelection
from PartSegCore.segmentation import ROIExtractionResult

if typing.TYPE_CHECKING:
    from qtpy.QtGui import QHideEvent, QShowEvent  # pragma: no cover
SELECT_TEXT = "<select>"


class NapariInteractiveAlgorithmSettingsWidget(InteractiveAlgorithmSettingsWidget):
    form_widget: NapariFormWidgetWithMask

    def _form_widget(self, algorithm, start_values) -> FormWidget:
        return NapariFormWidgetWithMask(
            algorithm.__argument_class__ if algorithm.__new_style__ else algorithm.get_fields(),
            start_values=start_values,
            parent=self,
        )

    def reset_choices(self, event=None):
        self.form_widget.reset_choices(event)

    def get_layer_list(self) -> typing.List[str]:
        return [x.name for x in self.get_layers().values() if x.name != "mask"]

    def get_layers(self) -> typing.Dict[str, Layer]:
        values = self.form_widget.get_layers()
        return {k: v for k, v in values.items() if isinstance(v, Layer)}


class NapariAlgorithmChoose(AlgorithmChooseBase):
    def _algorithm_widget(self, settings, val) -> InteractiveAlgorithmSettingsWidget:
        return NapariInteractiveAlgorithmSettingsWidget(settings, val, [], parent=self)

    def reset_choices(self, event=None):
        for widget in self.algorithm_dict.values():
            widget.reset_choices(event)


class ROIExtractionAlgorithms(QWidget):
    @staticmethod
    def get_method_dict() -> typing.Type[AlgorithmSelection]:  # pragma: no cover
        raise NotImplementedError

    @staticmethod
    def prefix() -> str:  # pragma: no cover
        raise NotImplementedError

    def __init__(self, napari_viewer: Viewer):
        plugins.register_if_need()
        super().__init__()
        self._scale = np.array((1, 1, 1))
        self.channel_names = []
        self.mask_name = ""

        self.viewer = napari_viewer
        self.settings = get_settings()
        self.algorithm_chose = NapariAlgorithmChoose(self.settings, self.get_method_dict())
        self.calculate_btn = QPushButton("Run")
        self.calculate_btn.clicked.connect(self._run_calculation)
        self.info_text = TextShow()

        self.profile_combo_box = SearchComboBox()
        self.profile_combo_box.addItem(SELECT_TEXT)
        self.profile_combo_box.addItems(list(self.profile_dict.keys()))
        self.save_btn = QPushButton("Save parameters")
        self.manage_btn = QPushButton("Manage parameters")
        self.target_layer_name = QLineEdit()
        self.target_layer_name.setText(self.settings.get(f"{self.prefix()}.target_layer_name", "ROI"))

        layout = QVBoxLayout()
        btn_layout = QHBoxLayout()
        btn_layout.addWidget(self.save_btn)
        btn_layout.addWidget(self.manage_btn)
        layout.addLayout(btn_layout)
        target_layer_layout = QHBoxLayout()
        target_layer_layout.addWidget(QLabel("Target layer name:"))
        target_layer_layout.addWidget(self.target_layer_name)
        layout.addLayout(target_layer_layout)
        layout.addWidget(self.profile_combo_box)
        layout.addWidget(self.calculate_btn)
        layout.addWidget(self.algorithm_chose, 1)
        layout.addWidget(self.info_text)

        self.setLayout(layout)

        self.algorithm_chose.result.connect(self.set_result)
        self.algorithm_chose.finished.connect(self._enable_calculation_btn)
        self.algorithm_chose.algorithm_changed.connect(self.algorithm_changed)
        self.save_btn.clicked.connect(self.save_action)
        self.manage_btn.clicked.connect(self.manage_action)
        self.profile_combo_box.textActivated.connect(self.select_profile)

        self.update_tooltips()
        register_plugins()

    def _enable_calculation_btn(self):
        self.calculate_btn.setEnabled(True)

    def manage_action(self):
        dialog = ProfilePreviewDialog(self.profile_dict, self.get_method_dict(), self.settings, parent=self)
        dialog.exec_()
        self.refresh_profiles()

    def select_profile(self, text):
        if text in [SELECT_TEXT, ""]:
            return  # pragma: no cover
        profile = self.profile_dict[text]
        self.algorithm_chose.change_algorithm(profile.algorithm, profile.values)
        self.profile_combo_box.setCurrentIndex(0)

    @property
    def profile_dict(self) -> typing.Dict[str, ROIExtractionProfile]:
        return self.settings.get_from_profile(f"{self.prefix()}.profiles", {})

    def save_action(self):
        widget = typing.cast(NapariInteractiveAlgorithmSettingsWidget, self.algorithm_chose.current_widget())
        profiles = self.profile_dict
        while True:
            text, ok = QInputDialog.getText(self, "Profile Name", "Input profile name here")
            if not ok:
                return  # pragma: no cover
            if text not in profiles or QMessageBox.Yes == QMessageBox.warning(
                self,
                "Already exists",
                "Profile with this name already exist. Overwrite?",
                QMessageBox.Yes | QMessageBox.No,
                QMessageBox.No,
            ):
                break  # pragma: no cover
        resp = ROIExtractionProfile(name=text, algorithm=widget.name, values=widget.get_values())
        profiles[text] = resp
        self.settings.dump()
        self.profile_combo_box.addItem(text)
        self.update_tooltips()

    def update_tooltips(self):
        for i in range(1, self.profile_combo_box.count()):
            if self.profile_combo_box.itemData(i, Qt.ToolTipRole) is not None:
                continue
            text = self.profile_combo_box.itemText(i)
            profile: ROIExtractionProfile = self.profile_dict[text]
            tool_tip_text = str(profile)
            self.profile_combo_box.setItemData(i, tool_tip_text, Qt.ToolTipRole)

    def algorithm_changed(self):
        self._scale = np.array((1, 1, 1))
        self.channel_names = []
        self.mask_name = ""

    def update_mask(self):
        widget = typing.cast(NapariInteractiveAlgorithmSettingsWidget, self.algorithm_chose.current_widget())
        mask = widget.get_layers().get("mask", None)
        if getattr(mask, "name", "") != self.mask_name or (widget.mask() is None and mask is not None):
            widget.set_mask(getattr(mask, "data", None))
            self.mask_name = getattr(mask, "name", "")

    def update_image(self):
        widget = typing.cast(NapariInteractiveAlgorithmSettingsWidget, self.algorithm_chose.current_widget())
        self.settings.last_executed_algorithm = widget.name
        layer_names: typing.List[str] = widget.get_layer_list()
        if layer_names == self.channel_names:
            return
        image = generate_image(self.viewer, *layer_names)

        self._scale = np.array(image.spacing)
        self.channel_names = image.channel_names
        widget.image_changed(image)
        self.mask_name = ""

    def _run_calculation(self):
        widget = typing.cast(NapariInteractiveAlgorithmSettingsWidget, self.algorithm_chose.current_widget())
        self.settings.last_executed_algorithm = widget.name
        self.update_image()
        self.update_mask()
        widget.execute()
        self.calculate_btn.setDisabled(True)

    def showEvent(self, event: "QShowEvent") -> None:
        self.reset_choices(None)
        super().showEvent(event)

    def hideEvent(self, event: "QHideEvent") -> None:
        self.settings.dump()
        super().hideEvent(event)

    def reset_choices(self, event=None):
        self.algorithm_chose.reset_choices(event)

    def set_result(self, result: ROIExtractionResult):
        if result.info_text:
            show_info(result.info_text)
        if len(result.roi_info.bound_info) == 0:
            if not result.info_text:
                show_info("There is no ROI in result. Please check algorithm parameters.")
            return
        roi = result.roi
        if self.sender() is not None:
            self.info_text.setPlainText(self.sender().get_info_text())
            with suppress(Exception):
                roi = self.sender().current_widget().algorithm_thread.algorithm.image.fit_array_to_image(result.roi)

        layer_name = self.target_layer_name.text()
        self.settings.set(f"{self.prefix()}.target_layer_name", layer_name)
        column_list = []
        column_set = set()
        for value in result.roi_annotation.values():
            for column_name in value.items():
                if column_name not in column_set:
                    column_list.append(column_name)
                    column_set.add(column_name)
        properties = pd.DataFrame.from_dict(result.roi_annotation, orient="index")
        properties["index"] = list(result.roi_annotation.keys())
        if layer_name in self.viewer.layers:
            self.viewer.layers[layer_name].data = result.roi
            self.viewer.layers[layer_name].metadata = {"parameters": result.parameters}
            self.viewer.layers[layer_name].properties = properties
        else:
            self.viewer.add_labels(
                roi,
                scale=np.array(self._scale)[-result.roi.ndim :] * UNIT_SCALE[Units.nm.value],
                name=layer_name,
                metadata={"parameters": result.parameters},
                properties=properties,
            )

    def refresh_profiles(self):
        self.profile_combo_box.clear()
        self.profile_combo_box.addItem(SELECT_TEXT)
        self.profile_combo_box.addItems(list(self.profile_dict.keys()))


class ROIAnalysisExtraction(ROIExtractionAlgorithms):
    @staticmethod
    def get_method_dict():
        return AnalysisAlgorithmSelection

    @staticmethod
    def prefix() -> str:
        return "roi_analysis_extraction"


class ROIMaskExtraction(ROIExtractionAlgorithms):
    @staticmethod
    def get_method_dict():
        return MaskAlgorithmSelection

    @staticmethod
    def prefix() -> str:
        return "roi_mask_extraction"


class ProfilePreviewDialog(QDialog):
    def __init__(
        self,
        profile_dict: typing.Dict[str, ROIExtractionProfile],
        algorithm_selection: typing.Type[AlgorithmSelection],
        settings: BaseSettings,
        parent=None,
    ):
        super().__init__(parent=parent)
        self.profile_dict = profile_dict
        self.algorithm_selection = algorithm_selection
        self.settings = settings

        self.profile_list = SearchableListWidget()
        self.profile_list.addItems(list(self.profile_dict.keys()))
        self.profile_list.currentTextChanged.connect(self.profile_selected)
        self.profile_view = QPlainTextEdit()
        self.profile_view.setReadOnly(True)

        self.delete_btn = QPushButton("Delete")
        self.delete_btn.clicked.connect(self.delete_action)
        self.rename_btn = QPushButton("Rename")
        self.rename_btn.clicked.connect(self.rename_action)
        self.export_btn = QPushButton("Export")
        self.export_btn.clicked.connect(self.export_action)
        self.import_btn = QPushButton("Import")
        self.import_btn.clicked.connect(self.import_action)
        layout = QGridLayout()
        layout.addWidget(self.profile_list, 0, 0)
        layout.addWidget(self.profile_view, 0, 1)
        layout.addWidget(self.delete_btn, 1, 0)
        layout.addWidget(self.rename_btn, 1, 1)
        layout.addWidget(self.export_btn, 2, 0)
        layout.addWidget(self.import_btn, 2, 1)

        self.setLayout(layout)

    def profile_selected(self, text):
        if text not in self.profile_dict:
            return
        profile = self.profile_dict[text]
        self.profile_view.setPlainText(str(profile))

    def delete_action(self):
        if self.profile_list.currentItem() is None:
            return  # pragma: no cover
        if self.profile_list.currentItem().text() in self.profile_dict:
            del self.profile_dict[self.profile_list.currentItem().text()]
        self.profile_list.clear()
        self.profile_list.addItems(list(self.profile_dict.keys()))

    def rename_action(self):
        if self.profile_list.currentItem() is None:
            return  # pragma: no cover
        old_name = self.profile_list.currentItem().text()
        if old_name not in self.profile_dict:
            return  # pragma: no cover

        text, ok = QInputDialog.getText(self, "Profile Name", "Input profile name here")
        if not ok:
            return  # pragma: no cover

        if text in self.profile_dict:  # pragma: no cover
            QMessageBox.warning(
                self,
                "Already exists",
                "Profile with this name already exist.",
                QMessageBox.Ok,
                QMessageBox.Ok,
            )
            self.rename_action()
            return
        profile = self.profile_dict[old_name]
        del self.profile_dict[old_name]
        profile.name = text
        self.profile_dict[text] = profile
        self.profile_list.clear()
        self.profile_list.addItems(list(self.profile_dict.keys()))

    def export_action(self):
        exp = ExportDialog(self.profile_dict, ProfileDictViewer, parent=self)
        if not exp.exec_():
            return  # pragma: no cover
        dial = PSaveDialog(SaveProfilesToJSON, settings=self.settings, parent=self, path=IO_SAVE_DIRECTORY)
        if dial.exec_():
            save_location, _selected_filter, save_class, values = dial.get_result()
            data = {x: self.profile_dict[x] for x in exp.get_export_list()}
            save_class.save(save_location, data, values)

    def import_action(self):
        dial = PLoadDialog(LoadProfileFromJSON, settings=self.settings, parent=self, path=IO_SAVE_DIRECTORY)
        if not dial.exec_():
            return  # pragma: no cover
        file_list, _, load_class = dial.get_result()
        profs, err = load_class.load(file_list)
        if err:
            show_warning("Import error", "error during importing, part of data were filtered.")  # pragma: no cover
        imp = ImportDialog(profs, self.profile_dict, ProfileDictViewer, parent=self)
        if not imp.exec_():
            return  # pragma: no cover
        for original_name, final_name in imp.get_import_list():
            self.profile_dict[final_name] = profs[original_name]
        self.settings.dump()
        self.profile_list.clear()
        self.profile_list.addItems(list(self.profile_dict.keys()))



import importlib
import os
import warnings
from functools import partial
from typing import TYPE_CHECKING, Type

from qtpy.QtCore import QSize, Qt, QThread, Signal
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QGridLayout, QMainWindow, QMessageBox, QProgressBar, QToolButton, QWidget

from PartSeg import ANALYSIS_NAME, APP_NAME, MASK_NAME, state_store
from PartSeg.common_backend import napari_get_settings
from PartSeg.common_backend.base_settings import BaseSettings, get_stylesheet, get_theme, napari_template
from PartSeg.common_backend.load_backup import import_config
from PartSegData import icons_dir
from PartSegImage import TiffImageReader

if TYPE_CHECKING:
    from PartSeg.common_gui.main_window import BaseMainWindow


class Prepare(QThread):
    def __init__(self, module: str):
        super().__init__()
        self.module = module
        self.result = None
        self.errors = []

    def run(self):
        if not self.module:  # pragma: no cover
            return

        from PartSeg import plugins

        plugins.register()
        main_window_module = importlib.import_module(self.module)
        main_window: Type[BaseMainWindow] = main_window_module.MainWindow
        settings: BaseSettings = main_window.get_setting_class()(main_window_module.CONFIG_FOLDER)
        self.errors = settings.load()
        reader = TiffImageReader()
        im = reader.read(main_window.initial_image_path)
        im.file_path = ""
        self.result = partial(main_window, settings=settings, initial_image=im)


class PartSegGUILauncher(QWidget):
    launched = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.lib_path = ""
        self.final_title = ""
        analysis_icon = QIcon(os.path.join(icons_dir, "icon.png"))
        stack_icon = QIcon(os.path.join(icons_dir, "icon_stack.png"))
        self.analysis_button = QToolButton(self)
        self.analysis_button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        self.analysis_button.setIcon(analysis_icon)
        # TODO use more general solution for text wrapping
        self.analysis_button.setText(ANALYSIS_NAME.replace(" ", "\n"))
        self.analysis_button.setIconSize(QSize(100, 100))
        self.mask_button = QToolButton(self)
        self.mask_button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        self.mask_button.setIcon(stack_icon)
        self.mask_button.setText(MASK_NAME.replace(" ", "\n"))
        self.mask_button.setIconSize(QSize(100, 100))
        self.analysis_button.clicked.connect(self.launch_analysis)
        self.mask_button.clicked.connect(self.launch_mask)
        self.progress = QProgressBar()
        self.progress.setHidden(True)
        layout = QGridLayout()
        layout.addWidget(self.progress, 0, 0, 1, 2)
        layout.addWidget(self.analysis_button, 1, 1)
        layout.addWidget(self.mask_button, 1, 0)
        self.setLayout(layout)
        self.prepare = Prepare("")
        self.prepare.finished.connect(self.launch)
        self.wind = []

    def launch_analysis(self):
        self._launch_begin()
        self._launch_analysis()
        self.prepare.start()

    def _launch_analysis(self):
        self.lib_path = "PartSeg._roi_analysis.main_window"
        self.final_title = f"{APP_NAME} {ANALYSIS_NAME}"
        self.prepare.module = self.lib_path

    def launch_mask(self):
        self._launch_begin()
        self._launch_mask()
        self.prepare.start()

    def _launch_mask(self):
        self.lib_path = "PartSeg._roi_mask.main_window"
        self.final_title = f"{APP_NAME} {MASK_NAME}"
        self.prepare.module = self.lib_path

    def _launch_begin(self):
        self.progress.setVisible(True)
        self.progress.setRange(0, 0)
        self.analysis_button.setDisabled(True)
        self.mask_button.setDisabled(True)
        import_config()

    def window_shown(self):
        self.progress.setHidden(True)
        self.analysis_button.setEnabled(True)
        self.mask_button.setEnabled(True)
        self.launched.emit()

    def launch(self):
        if self.prepare.result is None:  # pragma: no cover
            self.close()
            return
        if self.prepare.errors:  # pragma: no cover
            errors_message = QMessageBox()
            errors_message.setText("There are errors during start")
            errors_message.setInformativeText(
                "During load saved state some of data could not be load properly\n"
                "The files has prepared backup copies in  state directory (Help > State directory)"
            )
            errors_message.setStandardButtons(QMessageBox.Ok)
            text = "\n".join(f"File: {x[0]}\n{x[1]}" for x in self.prepare.errors)

            errors_message.setDetailedText(text)
            errors_message.exec_()
        wind = self.prepare.result(title=self.final_title, signal_fun=self.window_shown)
        wind.show()
        self.wind.append(wind)


class MainWindow(QMainWindow):
    def __init__(self, title):
        super().__init__()
        self.setWindowTitle(title)
        widget = PartSegGUILauncher(self)
        widget.launched.connect(self.close)
        self.setCentralWidget(widget)
        self.setWindowIcon(QIcon(os.path.join(icons_dir, "icon.png")))
        self._update_theme()

    def _update_theme(self):
        napari_settings = napari_get_settings(state_store.save_folder)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", FutureWarning)
            theme = get_theme(napari_settings.appearance.theme, as_dict=True)
        # TODO understand qss overwrite mechanism
        self.setStyleSheet(napari_template(get_stylesheet(), **theme))



import pytest

from PartSeg._roi_mask.batch_proceed import BatchProceed, BatchTask
from PartSegCore.mask.io_functions import SaveROIOptions


@pytest.fixture()
def batch_task(stack_image, mask_threshold_profile):
    return BatchTask(stack_image, mask_threshold_profile, None)


class TestBatchProceed:
    def test_create(self, qapp):
        BatchProceed()

    def test_add_task(self, qapp, batch_task):
        thread = BatchProceed()
        thread.add_task(batch_task)
        assert thread.queue.qsize() == 1
        thread.add_task(batch_task)
        assert thread.queue.qsize() == 2
        thread.add_task([batch_task, batch_task])
        assert thread.queue.qsize() == 4

    def test_simple_run_task(self, qtbot, batch_task):
        thread = BatchProceed()
        thread.add_task(batch_task)
        assert thread.index == 0
        with qtbot.waitSignal(thread.multiple_result):
            thread.run_calculation()

    def test_run_with_save(self, qtbot, stack_image, mask_threshold_profile, tmp_path):
        batch_task = BatchTask(stack_image, mask_threshold_profile, (tmp_path, SaveROIOptions()))
        thread = BatchProceed()
        thread.add_task([batch_task, batch_task, batch_task])
        assert thread.index == 0
        thread.run_calculation()

        assert (tmp_path / "test_path.seg").exists()
        assert (tmp_path / "test_path_version1.seg").exists()
        assert (tmp_path / "test_path_version2.seg").exists()



import os
from contextlib import suppress
from functools import partial
from typing import Type

import numpy as np
from qtpy.QtCore import QByteArray, Qt, Signal, Slot
from qtpy.QtGui import QCloseEvent, QGuiApplication, QIcon, QKeySequence, QTextOption
from qtpy.QtWidgets import (
    QAbstractSpinBox,
    QCheckBox,
    QDoubleSpinBox,
    QFormLayout,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QProgressBar,
    QPushButton,
    QSizePolicy,
    QSpinBox,
    QTabWidget,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)
from superqt import QEnumComboBox

import PartSegData
from PartSeg import state_store
from PartSeg._roi_mask.batch_proceed import BatchProceed, BatchTask
from PartSeg._roi_mask.image_view import StackImageView
from PartSeg._roi_mask.segmentation_info_dialog import SegmentationInfoDialog
from PartSeg._roi_mask.simple_measurements import SimpleMeasurements
from PartSeg._roi_mask.stack_settings import StackSettings, get_mask
from PartSeg.common_backend.base_settings import IO_SAVE_DIRECTORY, ROI_NOT_FIT
from PartSeg.common_gui.advanced_tabs import AdvancedWindow
from PartSeg.common_gui.algorithms_description import AlgorithmChoose, InteractiveAlgorithmSettingsWidget
from PartSeg.common_gui.channel_control import ChannelProperty
from PartSeg.common_gui.custom_load_dialog import PLoadDialog
from PartSeg.common_gui.custom_save_dialog import PSaveDialog
from PartSeg.common_gui.exception_hooks import load_data_exception_hook
from PartSeg.common_gui.flow_layout import FlowLayout
from PartSeg.common_gui.main_window import OPEN_DIRECTORY, OPEN_FILE, OPEN_FILE_FILTER, BaseMainMenu, BaseMainWindow
from PartSeg.common_gui.mask_widget import MaskDialogBase
from PartSeg.common_gui.multiple_file_widget import MultipleFileWidget
from PartSeg.common_gui.napari_image_view import LabelEnum
from PartSeg.common_gui.select_multiple_files import AddFiles
from PartSeg.common_gui.stack_image_view import ColorBar
from PartSeg.common_gui.universal_gui_part import right_label
from PartSeg.common_gui.waiting_dialog import ExecuteFunctionDialog
from PartSegCore import UNIT_SCALE, Units
from PartSegCore.io_utils import WrongFileTypeException
from PartSegCore.mask import io_functions
from PartSegCore.mask.algorithm_description import MaskAlgorithmSelection
from PartSegCore.mask.history_utils import create_history_element_from_segmentation_tuple
from PartSegCore.mask.io_functions import LoadROI, LoadROIFromTIFF, LoadROIParameters, MaskProjectTuple, SaveROI
from PartSegCore.project_info import HistoryElement, HistoryProblem, calculate_mask_from_project
from PartSegCore.roi_info import ROIInfo
from PartSegImage import Image, TiffImageReader

CONFIG_FOLDER = os.path.join(state_store.save_folder, "mask")


class MaskDialog(MaskDialogBase):
    # FIXME add tests
    def __init__(self, settings: StackSettings):
        super().__init__(settings)
        self.settings = settings

    def next_mask(self):
        project_info: MaskProjectTuple = self.settings.get_project_info()
        mask_property = self.mask_widget.get_mask_property()
        self.settings.set("mask_manager.mask_property", mask_property)
        mask = calculate_mask_from_project(mask_description=mask_property, project=project_info)

        self.settings.add_history_element(
            create_history_element_from_segmentation_tuple(
                project_info,
                mask_property,
            )
        )
        self.settings.mask = mask
        self.settings.chosen_components_widget.un_check_all()
        self.close()

    def prev_mask(self):
        history: HistoryElement = self.settings.history_pop()
        history.arrays.seek(0)
        seg = np.load(history.arrays)
        history.arrays.seek(0)
        self.settings._set_roi_info(  # pylint: disable=protected-access
            ROIInfo(seg["segmentation"]),
            False,
            history.roi_extraction_parameters["selected"],
            history.roi_extraction_parameters["parameters"],
        )
        self.settings.mask = seg.get("mask")
        self.close()


class MainMenu(BaseMainMenu):
    image_loaded = Signal()

    def __init__(self, settings: StackSettings, main_window):
        """
        :type settings: StackSettings
        :param settings:
        """
        super().__init__(settings, main_window)
        self.settings = settings
        self.segmentation_cache = None
        self.read_thread = None
        self.advanced_window = None
        self.measurements_window = None
        self.load_image_btn = QPushButton("Load image")
        self.load_image_btn.clicked.connect(self.load_image)
        self.load_segmentation_btn = QPushButton("Load segmentation")
        self.load_segmentation_btn.clicked.connect(self.load_segmentation)
        self.save_segmentation_btn = QPushButton("Save segmentation")
        self.save_segmentation_btn.clicked.connect(self.save_segmentation)
        self.save_catted_parts = QPushButton("Save components")
        self.save_catted_parts.clicked.connect(self.save_result)
        self.advanced_window_btn = QPushButton("Advanced settings")
        self.advanced_window_btn.clicked.connect(self.show_advanced_window)
        self.mask_manager_btn = QPushButton("Mask manager")
        self.mask_manager_btn.clicked.connect(self.mask_manager)
        self.measurements_btn = QPushButton("Simple measurements")
        self.measurements_btn.clicked.connect(self.simple_measurement)
        self.segmentation_dialog = SegmentationInfoDialog(
            self.main_window.settings,
            self.main_window.options_panel.algorithm_options.algorithm_choose_widget.change_algorithm,
        )

        self.setContentsMargins(0, 0, 0, 0)
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.load_image_btn)
        layout.addWidget(self.load_segmentation_btn)
        layout.addWidget(self.save_catted_parts)
        layout.addWidget(self.save_segmentation_btn)
        layout.addWidget(self.advanced_window_btn)
        layout.addWidget(self.mask_manager_btn)
        layout.addWidget(self.measurements_btn)
        self.setLayout(layout)

    def simple_measurement(self):
        if self.measurements_window is None:
            self.measurements_window = SimpleMeasurements(self.settings)
        self.measurements_window.show()

    def mask_manager(self):
        if self.settings.roi is None:
            QMessageBox.information(self, "No segmentation", "Cannot create mask without segmentation")
            return
        if not self.settings.chosen_components():
            QMessageBox.information(self, "No selected components", "Mask is created only from selected components")
            return
        dial = MaskDialog(self.settings)
        dial.exec_()

    def show_advanced_window(self):
        if self.advanced_window is None:
            self.advanced_window = AdvancedWindow(self.settings, ["channelcontrol"], reload_list=[self.reload])
        self.advanced_window.show()

    def reload(self):
        self.parent().parent().options_panel.algorithm_options.algorithm_choose_widget.reload(MaskAlgorithmSelection)

    def load_image(self):
        # TODO move segmentation with image load to load_segmentaion
        dial = PLoadDialog(
            io_functions.load_dict,
            settings=self.settings,
            path=OPEN_DIRECTORY,
            filter_path=OPEN_FILE_FILTER,
        )
        default_file_path = self.settings.get(OPEN_FILE, "")
        if os.path.isfile(default_file_path):
            dial.selectFile(default_file_path)
        if not dial.exec_():
            return
        load_property = dial.get_result()
        self.settings.set(OPEN_FILE, load_property.load_location[0])
        self.settings.add_last_files(load_property.load_location, load_property.load_class.get_name())

        execute_dialog = ExecuteFunctionDialog(
            load_property.load_class.load,
            [load_property.load_location],
            {"metadata": {"default_spacing": self.settings.image.spacing}},
            text="Load data",
            exception_hook=load_data_exception_hook,
        )
        if execute_dialog.exec_():
            result = execute_dialog.get_result()
            if result is None:
                return
            self.set_data(result)

    def set_image(self, image: Image) -> bool:
        if image is None:
            return False
        if image.is_time:
            if image.is_stack:
                QMessageBox.warning(self, "Not supported", "Data that are time data are currently not supported")
                return False

            res = QMessageBox.question(
                self,
                "Not supported",
                "Time data are currently not supported. Maybe You would like to treat time as z-stack",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )

            if res == QMessageBox.StandardButton.Yes:
                image = image.swap_time_and_stack()
            else:
                return False
        self.settings.image = image
        return True

    def load_segmentation(self):
        settings_path = OPEN_DIRECTORY if self.settings.get("sync_dirs", False) else "io.open_segmentation_directory"
        dial = PLoadDialog(
            {
                LoadROI.get_name(): LoadROI,
                LoadROIParameters.get_name(): LoadROIParameters,
                LoadROIFromTIFF.get_name(): LoadROIFromTIFF,
            },
            settings=self.settings,
            path=settings_path,
        )
        if not dial.exec_():
            return
        load_property = dial.get_result()

        def exception_hook(exception):
            mess = QMessageBox(self)
            if isinstance(exception, ValueError) and exception.args[0] == "Segmentation do not fit to image":
                mess.warning(self, "Open error", "Segmentation do not fit to image")
            elif isinstance(exception, MemoryError):
                mess.warning(self, "Open error", "Not enough memory to read this image")
            elif isinstance(exception, IOError):
                mess.warning(self, "Open error", "Some problem with reading from disc")
            elif isinstance(exception, WrongFileTypeException):
                mess.warning(
                    self,
                    "Open error",
                    "No needed files inside archive. Most probably you choose file from segmentation analysis",
                )
            else:
                raise exception

        dial = ExecuteFunctionDialog(
            load_property.load_class.load,
            [load_property.load_location],
            text="Load segmentation",
            exception_hook=exception_hook,
        )
        if not dial.exec_():
            return
        result: MaskProjectTuple = dial.get_result()
        if result is None:
            QMessageBox.critical(self, "Data Load fail", "Fail of loading data")
            return
        if result.roi_info.roi is not None:
            try:
                self.settings.set_project_info(dial.get_result())
                return
            except ValueError as e:
                if e.args != (ROI_NOT_FIT,):
                    raise
                text = (
                    f"Segmentation of shape {result.roi_info.roi.shape}\n({result.file_path})\n"
                    f"do not fit to image of shape {self.settings.image.shape}\n({self.settings.image.file_path})"
                )
                if all(x is None for x in result.roi_extraction_parameters.values()):
                    QMessageBox.warning(self, "Segmentation do not fit", f"{text} and no parameters for extraction")
                    return

                self.segmentation_dialog.set_additional_text(text + "\nmaybe you would like to load parameters only.")
            except HistoryProblem:
                QMessageBox().warning(
                    self,
                    "Load Problem",
                    "You set to save selected components when loading "
                    "another segmentation but history is incompatibility",
                )

        else:
            self.segmentation_dialog.set_additional_text("")
        self.segmentation_dialog.set_parameters_dict(result.roi_extraction_parameters)
        self.segmentation_dialog.show()

    def save_segmentation(self):
        if self.settings.roi is None:
            QMessageBox.warning(self, "No segmentation", "No segmentation to save")
            return
        settings_path = OPEN_DIRECTORY if self.settings.get("sync_dirs", False) else "io.save_segmentation_directory"
        dial = PSaveDialog(
            io_functions.save_segmentation_dict,
            system_widget=False,
            settings=self.settings,
            path=settings_path,
        )

        dial.selectFile(f"{os.path.splitext(os.path.basename(self.settings.image_path))[0]}.seg")

        if not dial.exec_():
            return
        save_location, _selected_filter, save_class, values = dial.get_result()

        def exception_hook(exception):
            QMessageBox.critical(
                self, "Save error", f"Error on disc operation. Text: {exception}", QMessageBox.StandardButton.Ok
            )
            raise exception

        dial = ExecuteFunctionDialog(
            save_class.save,
            [save_location, self.settings.get_project_info(), values],
            text="Save segmentation",
            exception_hook=exception_hook,
        )
        dial.exec_()

    def save_result(self):
        if self.settings.image_path is not None and QMessageBox.StandardButton.Yes == QMessageBox.question(
            self,
            "Copy",
            "Copy name to clipboard?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            QMessageBox.StandardButton.Yes,
        ):
            clipboard = QGuiApplication.clipboard()
            clipboard.setText(os.path.splitext(os.path.basename(self.settings.image_path))[0])

        if self.settings.roi is None or len(self.settings.sizes) == 1:
            QMessageBox.warning(self, "No components", "No components to save")
            return
        dial = PSaveDialog(
            io_functions.save_components_dict,
            system_widget=False,
            settings=self.settings,
            file_mode=PSaveDialog.Directory,
            path="io.save_components_directory",
        )
        dial.selectFile(os.path.splitext(os.path.basename(self.settings.image_path))[0])
        if not dial.exec_():
            return
        res = dial.get_result()
        potential_names = self.settings.get_file_names_for_save_result(res.save_destination)
        conflict = [el for el in potential_names if os.path.exists(el)]

        if conflict:
            # TODO modify because of long lists
            conflict_str = "\n".join(conflict)
            if QMessageBox.StandardButton.No == QMessageBox.warning(
                self,
                "Overwrite",
                f"Overwrite files:\n {conflict_str}",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            ):
                self.save_result()

        def exception_hook(exception):
            QMessageBox.critical(
                self, "Save error", f"Error on disc operation. Text: {exception}", QMessageBox.StandardButton.Ok
            )

        dial = ExecuteFunctionDialog(
            res.save_class.save,
            [res.save_destination, self.settings.get_project_info(), res.parameters],
            text="Save components",
            exception_hook=exception_hook,
        )
        dial.exec_()


class ComponentCheckBox(QCheckBox):
    mouse_enter = Signal(int)
    mouse_leave = Signal(int)

    def __init__(self, number: int, parent=None):
        super().__init__(str(number), parent)
        self.number = number

    def enterEvent(self, _event):
        self.mouse_enter.emit(self.number)

    def leaveEvent(self, _event):
        self.mouse_leave.emit(self.number)


class ChosenComponents(QWidget):
    """
    :type check_box: dict[int, ComponentCheckBox]
    """

    check_change_signal = Signal()
    mouse_enter = Signal(int)
    mouse_leave = Signal(int)

    def __init__(self):
        super().__init__()
        self.check_box = {}
        self.check_all_btn = QPushButton("Select all")
        self.check_all_btn.clicked.connect(self.check_all)
        self.un_check_all_btn = QPushButton("Unselect all")
        self.un_check_all_btn.clicked.connect(self.un_check_all)
        main_layout = QVBoxLayout()
        btn_layout = QHBoxLayout()
        btn_layout.addWidget(self.check_all_btn)
        btn_layout.addWidget(self.un_check_all_btn)
        self.check_layout = FlowLayout()
        main_layout.addLayout(btn_layout)
        main_layout.addLayout(self.check_layout)
        self.setLayout(main_layout)

    def other_component_choose(self, num):
        check = self.check_box[num]
        check.setChecked(not check.isChecked())

    def check_all(self):
        for el in self.check_box.values():
            el.setChecked(True)

    def un_check_all(self):
        for el in self.check_box.values():
            el.setChecked(False)

    def remove_components(self):
        self.check_layout.clear()
        for el in self.check_box.values():
            el.deleteLater()
            el.stateChanged.disconnect()
            el.mouse_leave.disconnect()
            el.mouse_enter.disconnect()
        self.check_box.clear()

    def new_choose(self, num, chosen_components):
        self.set_chose(range(1, num + 1), chosen_components)

    def set_chose(self, components_index, chosen_components):
        chosen_components = set(chosen_components)
        self.blockSignals(True)
        self.remove_components()
        for el in components_index:
            check = ComponentCheckBox(el)
            if el in chosen_components:
                check.setChecked(True)
            check.stateChanged.connect(self.check_change)
            check.mouse_enter.connect(self.mouse_enter.emit)
            check.mouse_leave.connect(self.mouse_leave.emit)
            self.check_box[el] = check
            self.check_layout.addWidget(check)
        self.blockSignals(False)
        self.update()
        self.check_change_signal.emit()

    def check_change(self):
        self.check_change_signal.emit()

    def change_state(self, num, val):
        self.check_box[num].setChecked(val)

    def get_state(self, num: int) -> bool:
        # TODO Check what situation create report of id ID: af9b57f074264169b4353aa1e61d8bc2
        return False if num >= len(self.check_box) else self.check_box[num].isChecked()

    def get_chosen(self):
        return [num for num, check in self.check_box.items() if check.isChecked()]

    def get_mask(self):
        res = [0]
        res.extend(check.isChecked() for _, check in sorted(self.check_box.items()))
        return np.array(res, dtype=np.uint8)


class AlgorithmOptions(QWidget):
    def __init__(self, settings: StackSettings, image_view: StackImageView):  # noqa: PLR0915
        super().__init__()
        self.settings = settings
        self.view_name = image_view.name
        self.show_result = QEnumComboBox(enum_class=LabelEnum)  # QCheckBox("Show result")
        self._set_show_label_from_settings()
        self.opacity = QDoubleSpinBox()
        self.opacity.setRange(0, 1)
        self.opacity.setSingleStep(0.1)
        self._set_opacity_from_settings()
        self.only_borders = QCheckBox("Only borders")
        self._set_border_mode_from_settings()
        self.borders_thick = QSpinBox()
        self.borders_thick.setRange(1, 25)
        self.borders_thick.setSingleStep(1)
        self._set_border_thick_from_settings()
        self.execute_in_background_btn = QPushButton("Execute in background")
        self.execute_in_background_btn.setToolTip("Run calculation in background. Put result in multiple files panel")
        self.execute_btn = QPushButton("Execute")
        self.execute_btn.setStyleSheet("QPushButton{font-weight: bold;}")
        self.execute_all_btn = QPushButton("Execute all")
        self.execute_all_btn.setToolTip(
            "Execute in batch mode segmentation with current parameter. File list need to be specified in image tab."
        )
        self.execute_all_btn.setDisabled(True)
        self.save_parameters_btn = QPushButton("Save parameters")
        self.block_execute_all_btn = False
        self.algorithm_choose_widget = AlgorithmChoose(settings, MaskAlgorithmSelection)
        self.algorithm_choose_widget.result.connect(self.execution_result_set)
        self.algorithm_choose_widget.finished.connect(self.execution_finished)
        self.algorithm_choose_widget.progress_signal.connect(self.progress_info)

        self.keep_chosen_components_chk = QCheckBox("Save selected components")
        self.keep_chosen_components_chk.setToolTip(
            "Save chosen components when loading segmentation form file\n or from multiple file widget."
        )
        self.keep_chosen_components_chk.stateChanged.connect(self.set_keep_chosen_components)
        self.keep_chosen_components_chk.setChecked(settings.keep_chosen_components)
        self.show_parameters = QPushButton("Show parameters")
        self.show_parameters.setToolTip("Show parameters of segmentation for each components")
        self.show_parameters_widget = SegmentationInfoDialog(
            self.settings, self.algorithm_choose_widget.change_algorithm
        )
        self.show_parameters.clicked.connect(self.show_parameters_widget.show)
        self.choose_components = ChosenComponents()
        self.choose_components.check_change_signal.connect(image_view.refresh_selected)
        self.choose_components.mouse_leave.connect(image_view.component_unmark)
        self.choose_components.mouse_enter.connect(image_view.component_mark)
        self.chosen_list = []
        self.progress_bar2 = QProgressBar()
        self.progress_bar2.setHidden(True)
        self.progress_bar = QProgressBar()
        self.progress_bar.setHidden(True)
        self.progress_info_lab = QLabel()
        self.progress_info_lab.setHidden(True)
        self.file_list = []
        self.batch_process = BatchProceed()
        self.batch_process.progress_signal.connect(self.progress_info)
        self.batch_process.error_signal.connect(self.execution_all_error)
        self.batch_process.execution_done.connect(self.execution_all_done)
        self.batch_process.range_signal.connect(self.progress_bar.setRange)
        self.is_batch_process = False

        self.setContentsMargins(0, 0, 0, 0)
        main_layout = QVBoxLayout()
        opt_layout = QHBoxLayout()
        opt_layout.setContentsMargins(0, 0, 0, 0)
        opt_layout.addWidget(self.show_result)
        opt_layout.addWidget(right_label("Opacity:"))
        opt_layout.addWidget(self.opacity)
        main_layout.addLayout(opt_layout)
        opt_layout2 = QHBoxLayout()
        opt_layout2.setContentsMargins(0, 0, 0, 0)
        opt_layout2.addWidget(self.only_borders)
        opt_layout2.addWidget(right_label("Border thick:"))
        opt_layout2.addWidget(self.borders_thick)
        main_layout.addLayout(opt_layout2)
        btn_layout = QGridLayout()
        btn_layout.setContentsMargins(0, 0, 0, 0)
        btn_layout.addWidget(self.execute_btn, 0, 0)
        btn_layout.addWidget(self.execute_in_background_btn, 0, 1)
        btn_layout.addWidget(self.execute_all_btn, 1, 0)
        btn_layout.addWidget(self.save_parameters_btn, 1, 1)
        main_layout.addLayout(btn_layout)
        main_layout.addWidget(self.progress_bar2)
        main_layout.addWidget(self.progress_bar)
        main_layout.addWidget(self.progress_info_lab)
        main_layout.addWidget(self.algorithm_choose_widget, 1)
        main_layout.addWidget(self.choose_components)
        down_layout = QHBoxLayout()
        down_layout.addWidget(self.keep_chosen_components_chk)
        down_layout.addWidget(self.show_parameters)
        main_layout.addLayout(down_layout)
        main_layout.addStretch()
        main_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(main_layout)

        # noinspection PyUnresolvedReferences
        self.execute_in_background_btn.clicked.connect(self.execute_in_background)
        self.execute_btn.clicked.connect(self.execute_action)
        self.execute_all_btn.clicked.connect(self.execute_all_action)
        self.save_parameters_btn.clicked.connect(self.save_parameters)
        # noinspection PyUnresolvedReferences
        self.opacity.valueChanged.connect(self._set_opacity)
        # noinspection PyUnresolvedReferences
        self.show_result.currentEnumChanged.connect(self._set_show_label)
        self.only_borders.stateChanged.connect(self._set_border_mode)
        # noinspection PyUnresolvedReferences
        self.borders_thick.valueChanged.connect(self._set_border_thick)
        image_view.component_clicked.connect(self.choose_components.other_component_choose)
        settings.chosen_components_widget = self.choose_components
        settings.components_change_list.connect(self.choose_components.new_choose)
        settings.image_changed.connect(self.choose_components.remove_components)
        settings.connect_to_profile(f"{self.view_name}.image_state.only_border", self._set_border_mode_from_settings)
        settings.connect_to_profile(f"{self.view_name}.image_state.border_thick", self._set_border_thick_from_settings)
        settings.connect_to_profile(f"{self.view_name}.image_state.opacity", self._set_opacity_from_settings)
        settings.connect_to_profile(f"{self.view_name}.image_state.show_label", self._set_show_label_from_settings)

    def _set_border_mode(self, value: bool):
        self.settings.set_in_profile(f"{self.view_name}.image_state.only_border", value)

    def _set_border_thick(self, value: int):
        self.settings.set_in_profile(f"{self.view_name}.image_state.border_thick", value)

    def _set_opacity(self, value: float):
        self.settings.set_in_profile(f"{self.view_name}.image_state.opacity", value)

    def _set_show_label(self, value: LabelEnum):
        self.settings.set_in_profile(f"{self.view_name}.image_state.show_label", value)

    def _set_border_mode_from_settings(self):
        self.only_borders.setChecked(self.settings.get_from_profile(f"{self.view_name}.image_state.only_border", True))

    def _set_border_thick_from_settings(self):
        self.borders_thick.setValue(self.settings.get_from_profile(f"{self.view_name}.image_state.border_thick", 1))

    def _set_opacity_from_settings(self):
        self.opacity.setValue(self.settings.get_from_profile(f"{self.view_name}.image_state.opacity", 1.0))

    def _set_show_label_from_settings(self):
        self.show_result.setCurrentEnum(
            self.settings.get_from_profile(f"{self.view_name}.image_state.show_label", LabelEnum.Show_results)
        )

    @Slot(int)
    def set_keep_chosen_components(self, val):
        self.settings.set_keep_chosen_components(val)

    def save_parameters(self):
        dial = PSaveDialog(
            io_functions.save_parameters_dict, system_widget=False, settings=self.settings, path=IO_SAVE_DIRECTORY
        )
        if not dial.exec_():
            return
        res = dial.get_result()
        res.save_class.save(
            save_location=res.save_destination,
            project_info=self.algorithm_choose_widget.current_parameters(),
        )

    def file_list_change(self, val):
        self.file_list = val
        if len(self.file_list) > 0 and not self.block_execute_all_btn:
            self.execute_all_btn.setEnabled(True)
        else:
            self.execute_all_btn.setDisabled(True)

    def get_chosen_components(self):
        return sorted(self.choose_components.get_chosen())

    @property
    def segmentation(self):
        return self.settings.roi

    @segmentation.setter
    def segmentation(self, val):
        self.settings.roi = val

    def _image_changed(self):
        self.settings.roi = None
        self.choose_components.set_chose([], [])

    def _execute_in_background_init(self):
        if self.batch_process.isRunning():
            return
        self.progress_bar2.setVisible(True)
        self.progress_bar2.setRange(0, self.batch_process.queue.qsize())
        self.progress_bar2.setValue(self.batch_process.index)
        self.progress_bar.setVisible(True)
        self.progress_bar.setValue(0)
        self.execute_btn.setDisabled(True)
        self.batch_process.start()

    def execute_in_background(self):
        # TODO check if components are properly passed
        widget = self.algorithm_choose_widget.current_widget()
        segmentation_profile = widget.get_segmentation_profile()
        task = BatchTask(self.settings.get_project_info(), segmentation_profile, None)
        self.batch_process.add_task(task)
        self.progress_bar2.setRange(0, self.progress_bar2.maximum() + 1)
        self._execute_in_background_init()

    def execute_all_action(self):
        dial = PSaveDialog(
            SaveROI,
            settings=self.settings,
            system_widget=False,
            path="io.save_batch",
            file_mode=PSaveDialog.Directory,
        )
        if not dial.exec_():
            return
        folder_path = str(dial.selectedFiles()[0])

        widget = self.algorithm_choose_widget.current_widget()

        save_parameters = dial.values
        segmentation_profile = widget.get_segmentation_profile()
        for file_path in self.file_list:
            task = BatchTask(file_path, segmentation_profile, (folder_path, save_parameters))
            self.batch_process.add_task(task)
        self.progress_bar2.setRange(0, self.progress_bar2.maximum() + len(self.file_list))
        self._execute_in_background_init()

    def execution_all_error(self, text):
        QMessageBox.warning(self, "Proceed error", text)

    def execution_all_done(self):
        if not self.batch_process.queue.empty():
            self._execute_in_background_init()
            return
        self.execute_btn.setEnabled(True)
        self.block_execute_all_btn = False
        if len(self.file_list) > 0:
            self.execute_all_btn.setEnabled(True)
        self.progress_bar.setHidden(True)
        self.progress_bar2.setHidden(True)
        self.progress_info_lab.setHidden(True)

    def execute_action(self):
        self.execute_btn.setDisabled(True)
        self.execute_all_btn.setDisabled(True)
        self.block_execute_all_btn = True
        self.is_batch_process = False
        self.progress_bar.setRange(0, 0)
        self.choose_components.setDisabled(True)
        chosen = sorted(self.choose_components.get_chosen())
        blank = get_mask(self.settings.roi, self.settings.mask, chosen)
        if blank is not None:
            # Problem with handling time data in algorithms
            # TODO Fix This
            blank = blank[0]
        self.progress_bar.setHidden(False)
        widget: InteractiveAlgorithmSettingsWidget = self.algorithm_choose_widget.current_widget()
        widget.set_mask(blank)
        self.progress_bar.setRange(0, widget.algorithm.get_steps_num())
        widget.execute()
        self.chosen_list = chosen

    def progress_info(self, text, num, file_name="", file_num=0):
        self.progress_info_lab.setVisible(True)
        if file_name:
            self.progress_info_lab.setText(file_name + "\n" + text)
        else:
            self.progress_info_lab.setText(text)
        self.progress_bar.setValue(num)
        self.progress_bar2.setValue(file_num)

    def execution_finished(self):
        self.execute_btn.setEnabled(True)
        self.block_execute_all_btn = False
        if len(self.file_list) > 0:
            self.execute_all_btn.setEnabled(True)
        self.progress_bar.setHidden(True)
        self.progress_info_lab.setHidden(True)
        self.choose_components.setDisabled(False)

    def execution_result_set(self, result):
        self.settings.set_segmentation_result(result)

    def showEvent(self, _):
        widget = self.algorithm_choose_widget.current_widget()
        widget.image_changed(self.settings.image)


class ImageInformation(QWidget):
    def __init__(self, settings: StackSettings, parent=None):
        """:type settings: ImageSettings"""
        super().__init__(parent)
        self._settings = settings
        self.path = QTextEdit("<b>Path:</b> example image")
        self.path.setWordWrapMode(QTextOption.WrapMode.WrapAnywhere)
        self.path.setReadOnly(True)
        self.setMinimumHeight(20)
        self.spacing = [QDoubleSpinBox() for _ in range(3)]
        self.multiple_files = QCheckBox("Show multiple files panel")
        self.multiple_files.setChecked(settings.get("multiple_files_widget", True))
        self.multiple_files.stateChanged.connect(self.set_multiple_files)
        self.sync_dirs = QCheckBox("Sync directories in file dialog")
        self.sync_dirs.setToolTip(
            "If checked then 'Load Image', 'Load segmentation' and 'Save segmentation' "
            "will open file dialog in the same directory"
        )
        self.sync_dirs.setChecked(settings.get("sync_dirs", False))
        self.sync_dirs.stateChanged.connect(self.set_sync_dirs)
        units_value = self._settings.get("units_value", Units.nm)
        for el in self.spacing:
            el.setAlignment(Qt.AlignmentFlag.AlignRight)
            el.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
            el.setRange(0, 100000)
            # noinspection PyUnresolvedReferences
            el.valueChanged.connect(self.image_spacing_change)
        self.units = QEnumComboBox(enum_class=Units)
        self.units.setCurrentEnum(units_value)
        # noinspection PyUnresolvedReferences
        self.units.currentIndexChanged.connect(self.update_spacing)

        self.add_files = AddFiles(settings, btn_layout=FlowLayout)

        spacing_layout = QFormLayout()
        spacing_layout.addRow("x:", self.spacing[0])
        spacing_layout.addRow("y:", self.spacing[1])
        spacing_layout.addRow("z:", self.spacing[2])
        spacing_layout.addRow("Units:", self.units)

        layout = QVBoxLayout()
        layout.addWidget(self.path)
        layout.addWidget(QLabel("Image spacing:"))
        layout.addLayout(spacing_layout)
        layout.addWidget(self.add_files)
        layout.addStretch(1)
        layout.addWidget(self.multiple_files)
        layout.addWidget(self.sync_dirs)
        self.setLayout(layout)
        self._settings.image_path_changed.connect(self.set_image_path)
        self._settings.connect_("multiple_files_widget", self._set_multiple_files)
        self._settings.connect_("sync_dirs", self._set_sync_dirs)

    @Slot(int)
    def set_multiple_files(self, val):
        self._settings.set("multiple_files_widget", val)

    @Slot(int)
    def set_sync_dirs(self, val):
        self._settings.set("sync_dirs", val)

    def _set_multiple_files(self):
        self.multiple_files.setChecked(self._settings.get("multiple_files_widget", True))

    def _set_sync_dirs(self):
        self.sync_dirs.setChecked(self._settings.get("sync_dirs", False))

    def update_spacing(self, index=None):
        units_value = self.units.currentEnum()
        if index is not None:
            self._settings.set("units_value", units_value)
        for el, val in zip(self.spacing, self._settings.image_spacing[::-1]):
            el.blockSignals(True)
            el.setValue(val * UNIT_SCALE[units_value.value])
            el.blockSignals(False)
        if self._settings.is_image_2d():
            self.spacing[2].setDisabled(True)
        else:
            self.spacing[2].setDisabled(False)

    def set_image_path(self, value):
        self.path.setText(f"<b>Path:</b> {value}")
        self.update_spacing()

    def image_spacing_change(self):
        self._settings.image_spacing = [el.value() / UNIT_SCALE[self.units.currentIndex()] for el in self.spacing[::-1]]

    def showEvent(self, _a0):
        units_value = self._settings.get("units_value", Units.nm)
        for el, val in zip(self.spacing, self._settings.image_spacing[::-1]):
            el.setValue(val * UNIT_SCALE[units_value.value])
        if self._settings.is_image_2d():
            self.spacing[2].setValue(0)
            self.spacing[2].setDisabled(True)
        else:
            self.spacing[2].setDisabled(False)


class Options(QTabWidget):
    def __init__(self, settings, image_view, parent=None):
        super().__init__(parent)
        self._settings = settings
        self.algorithm_options = AlgorithmOptions(settings, image_view)
        self.image_properties = ImageInformation(settings)
        self.image_properties.add_files.file_list_changed.connect(self.algorithm_options.file_list_change)
        self.algorithm_options.batch_process.multiple_result.connect(
            partial(self.image_properties.multiple_files.setChecked, True)
        )
        self.addTab(self.image_properties, "Image")
        self.addTab(self.algorithm_options, "Segmentation")
        self.setMinimumWidth(370)
        self.setCurrentIndex(1)

    def get_chosen_components(self):
        return self.algorithm_options.get_chosen_components()


class MainWindow(BaseMainWindow):
    settings: StackSettings

    @classmethod
    def get_setting_class(cls) -> Type[StackSettings]:
        return StackSettings

    initial_image_path = PartSegData.segmentation_mask_default_image

    def __init__(  # noqa: PLR0915
        self, config_folder=CONFIG_FOLDER, title="PartSeg", settings=None, signal_fun=None, initial_image=None
    ):
        super().__init__(config_folder, title, settings, io_functions.load_dict, signal_fun)
        self.channel_info = "channelcontrol"
        self.channel_control = ChannelProperty(self.settings, start_name="channelcontrol")
        self.image_view = StackImageView(self.settings, self.channel_control, name="channelcontrol")
        self.image_view.setMinimumWidth(450)
        self.info_text = QLabel()
        self.info_text.setSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Preferred)
        self.image_view.text_info_change.connect(self.info_text.setText)
        self.options_panel = Options(self.settings, self.image_view)
        self.main_menu = MainMenu(self.settings, self)
        self.main_menu.image_loaded.connect(self.image_read)
        self.settings.image_changed.connect(self.image_read)
        self.color_bar = ColorBar(self.settings, self.image_view)
        self.multiple_files = MultipleFileWidget(self.settings, io_functions.load_dict)
        self.multiple_files.setVisible(self.options_panel.image_properties.multiple_files.isChecked())
        self.options_panel.algorithm_options.batch_process.multiple_result.connect(
            partial(self.multiple_files.save_state_action, custom_name=False)
        )

        icon = QIcon(os.path.join(PartSegData.icons_dir, "icon_stack.png"))
        self.setWindowIcon(icon)

        menu_bar = self.menuBar()
        file_menu = menu_bar.addMenu("File")
        file_menu.addAction("&Open").triggered.connect(self.main_menu.load_image)
        file_menu.addMenu(self.recent_file_menu)
        file_menu.addAction("&Save segmentation").triggered.connect(self.main_menu.save_segmentation)
        file_menu.addAction("&Save components").triggered.connect(self.main_menu.save_result)
        view_menu = menu_bar.addMenu("View")
        view_menu.addAction("Settings and Measurement").triggered.connect(self.main_menu.show_advanced_window)
        view_menu.addAction("Additional output").triggered.connect(self.additional_layers_show)
        view_menu.addAction("Additional output with data").triggered.connect(lambda: self.additional_layers_show(True))
        view_menu.addAction("Napari viewer").triggered.connect(self.napari_viewer_show)
        view_menu.addAction("Toggle Multiple Files").triggered.connect(self.toggle_multiple_files)
        view_menu.addAction("Toggle console").triggered.connect(self._toggle_console)
        view_menu.addAction("Toggle scale bar").triggered.connect(self._toggle_scale_bar)
        action = view_menu.addAction("Screenshot")
        action.triggered.connect(self.screenshot(self.image_view))
        action.setShortcut(QKeySequence.StandardKey.Print)
        image_menu = menu_bar.addMenu("Image operations")
        image_menu.addAction("Image adjustment").triggered.connect(self.image_adjust_exec)
        help_menu = menu_bar.addMenu("Help")
        help_menu.addAction("State directory").triggered.connect(self.show_settings_directory)
        help_menu.addAction("About").triggered.connect(self.show_about_dialog)

        layout = QVBoxLayout()
        layout.addWidget(self.main_menu)
        sub_layout = QHBoxLayout()
        sub2_layout = QVBoxLayout()
        sub3_layout = QVBoxLayout()
        sub_layout.addWidget(self.multiple_files)
        sub_layout.addWidget(self.color_bar, 0)
        sub3_layout.addWidget(self.image_view, 1)
        sub3_layout.addWidget(self.info_text, 0)
        sub2_layout.addWidget(self.options_panel, 1)
        sub2_layout.addWidget(self.channel_control, 0)

        sub_layout.addLayout(sub3_layout, 1)
        sub_layout.addLayout(sub2_layout, 0)
        layout.addLayout(sub_layout)
        self.widget = QWidget()
        self.widget.setLayout(layout)
        self.setCentralWidget(self.widget)
        if initial_image is None:
            reader = TiffImageReader()
            im = reader.read(self.initial_image_path)
            im.file_path = ""
            self.settings.image = im
        elif initial_image is not False:
            self.settings.image = initial_image
        with suppress(KeyError):
            geometry = self.settings.get_from_profile("main_window_geometry")
            self.restoreGeometry(QByteArray.fromHex(bytes(geometry, "ascii")))

    def _toggle_scale_bar(self):
        self.image_view.toggle_scale_bar()
        super()._toggle_scale_bar()

    def closeEvent(self, event: QCloseEvent):
        self.settings.set_in_profile("main_window_geometry", self.saveGeometry().toHex().data().decode("ascii"))
        self.options_panel.algorithm_options.algorithm_choose_widget.recursive_get_values()
        self.main_menu.segmentation_dialog.close()
        self.options_panel.algorithm_options.show_parameters_widget.close()
        if self.main_menu.advanced_window is not None:
            self.main_menu.advanced_window.close()
            del self.main_menu.advanced_window
        if self.main_menu.measurements_window is not None:
            self.main_menu.measurements_window.close()
            del self.main_menu.measurements_window
        del self.main_menu.segmentation_dialog
        del self.options_panel.algorithm_options.show_parameters_widget
        super().closeEvent(event)

    @staticmethod
    def get_project_info(file_path, image, roi_info=None):
        if roi_info is None:
            roi_info = ROIInfo(None)
        return MaskProjectTuple(
            file_path=file_path,
            image=image,
            roi_info=roi_info,
            roi_extraction_parameters={i: None for i in roi_info.bound_info},
        )

    def set_data(self, data):
        self.main_menu.set_data(data)

    # @ensure_main_thread
    def change_theme(self, event):
        self.image_view.set_theme(self.settings.theme_name)
        super().change_theme(event)



import re
import typing
import warnings
from collections.abc import Iterable
from contextlib import suppress
from typing import Union

import numpy as np

from PartSegImage import Channel

Spacing = typing.Tuple[typing.Union[float, int], ...]
_IMAGE_DATA = typing.Union[typing.List[np.ndarray], np.ndarray]

_DEF = object()
FRAME_THICKNESS = 2

DEFAULT_SCALE_FACTOR = 10**9


def minimal_dtype(val: int):
    """
    Calculate minimal type to handle value in array

    :param val:
    :return: minimal dtype to handle given value
    :rtype:
    """
    if val < 250:
        return np.uint8
    return np.uint16 if val < 2**16 - 5 else np.uint32


def reduce_array(
    array: np.ndarray,
    components: typing.Optional[typing.Collection[int]] = None,
    max_val: typing.Optional[int] = None,
    dtype=None,
) -> np.ndarray:
    """
    Relabel components from 1 to components_num with keeping order

    :param array: array to relabel, deed to be integer type
    :param components: components to be keep, if None then all will be keep
    :param max_val: number of maximum component in array, if absent then will be calculated
        (to reduce whole array processing)
    :param dtype: type of returned array if no then minimal type is calculated
    :return: relabeled array in minimum type
    """
    # this function minimal dtype is np.uint8 so there is no need to do calculation.
    if components is None:
        components = np.unique(array.flat)
        if max_val is None:
            max_val = np.max(components)

    if max_val is None:
        max_val = np.max(array)

    translate = np.zeros(max_val + 1, dtype=dtype or minimal_dtype(len(components) + 1))

    for i, val in enumerate(sorted(components), start=0 if 0 in components else 1):
        translate[val] = i

    return translate[array]


class Image:
    """
    Base class for Images used in PartSeg

    :param data: 5-dim array with order: time, z, y, x, channel
    :param image_spacing: spacing for z, y, x
    :param file_path: path to image on disc
    :param mask: mask array in shape z,y,x
    :param default_coloring: default colormap - not used yet
    :param ranges: default ranges for channels
    :param channel_names: labels for channels
    :param axes_order: allow to create Image object form data with different axes order, or missed axes

    :cvar str ~.axis_order: internal order of axes

    It is prepared for subclassing with changed internal order. Eg:

    >>> class ImageJImage(Image):
    >>>     axis_order = "TZCYX"

    """

    _image_spacing: Spacing
    axis_order = "CTZYX"
    array_axis_order: str

    def __new__(cls, *args, **kwargs):
        if hasattr(cls, "return_order"):  # pragma: no cover
            warnings.warn("Using return_order is deprecated since PartSeg 0.11.0", DeprecationWarning, stacklevel=2)
            cls.axis_order = cls.return_order
        cls.array_axis_order = cls.axis_order.replace("C", "")
        return super().__new__(cls)

    def __init__(
        self,
        data: _IMAGE_DATA,
        image_spacing: Spacing,
        file_path=None,
        mask: typing.Union[None, np.ndarray] = None,
        default_coloring=None,
        ranges=None,
        channel_names=None,
        axes_order: typing.Optional[str] = None,
        shift: typing.Optional[Spacing] = None,
        name: str = "",
    ):
        # TODO add time distance to image spacing
        if axes_order is None:  # pragma: no cover
            warnings.warn(
                f"axes_order should be provided, Currently it uses {self.__class__}.axis_order",
                category=DeprecationWarning,
                stacklevel=2,
            )
            axes_order = self.axis_order
        if (isinstance(data, list) and any(x.ndim + 1 != len(axes_order) for x in data)) or (
            not isinstance(data, list) and data.ndim != len(axes_order)
        ):
            if isinstance(data, list):
                ndim = ", ".join([f"{x.ndim} + 1" for x in data])
            else:
                ndim = str(data.ndim)
            raise ValueError(
                "Data should have same number of dimensions "
                f"like length of axes_order (axis :{len(axes_order)}, ndim: {ndim}"
            )
        if not isinstance(image_spacing, tuple):
            image_spacing = tuple(image_spacing)
        self._channel_arrays = self._split_data_on_channels(data, axes_order)
        self._image_spacing = (1.0,) * (3 - len(image_spacing)) + image_spacing
        self._image_spacing = tuple(el if el > 0 else 10**-6 for el in self._image_spacing)

        self._shift = tuple(shift) if shift is not None else (0,) * len(self._image_spacing)
        self.name = name

        self.file_path = file_path
        self.default_coloring = default_coloring
        if self.default_coloring is not None:
            self.default_coloring = [np.array(x) for x in default_coloring]

        self._channel_names = self._prepare_channel_names(channel_names, self.channels)
        if ranges is None:
            self.ranges = list(
                zip((np.min(c) for c in self._channel_arrays), (np.max(c) for c in self._channel_arrays))
            )
        else:
            self.ranges = ranges
        self._mask_array = self._prepare_mask(mask, data, axes_order)
        if self._mask_array is not None:
            self._mask_array = self.fit_mask_to_image(self._mask_array)

    @classmethod
    def _prepare_mask(cls, mask, data, axes_order) -> typing.Optional[np.ndarray]:
        if mask is None:
            return None

        if isinstance(data, list):
            data_shape = list(data[0].shape)
        else:
            data_shape = list(data.shape)
            with suppress(ValueError):
                data_shape.pop(axes_order.index("C"))

        mask = cls._fit_array_to_image(data_shape, mask)
        return cls.reorder_axes(mask, axes_order.replace("C", ""))

    @staticmethod
    def _prepare_channel_names(channel_names, channels_num) -> typing.List[str]:
        default_channel_names = [f"channel {i + 1}" for i in range(channels_num)]
        if isinstance(channel_names, str):
            channel_names = [channel_names]
        if isinstance(channel_names, Iterable):
            channel_names_list = [str(x) for x in channel_names]
            channel_names_list = channel_names_list[:channels_num] + default_channel_names[len(channel_names_list) :]
        else:
            channel_names_list = default_channel_names
        return channel_names_list[:channels_num]

    @classmethod
    def _split_data_on_channels(
        cls, data: typing.Union[np.ndarray, typing.List[np.ndarray]], axes_order: str
    ) -> typing.List[np.ndarray]:
        if isinstance(data, list) and not axes_order.startswith("C"):  # pragma: no cover
            raise ValueError("When passing data as list of numpy arrays then Channel must be first axis.")
        if "C" not in axes_order:
            if not isinstance(data, np.ndarray):  # pragma: no cover
                raise TypeError("If `axes_order` does not contain `C` then data must be numpy array.")
            return [cls.reorder_axes(data, axes_order)]
        if axes_order.startswith("C"):
            if isinstance(data, list):
                dtype = np.result_type(*data)
                return [cls.reorder_axes(x, axes_order[1:]).astype(dtype) for x in data]
            return [cls.reorder_axes(x, axes_order[1:]) for x in data]

        if not isinstance(data, np.ndarray):
            raise TypeError("If `data` is list of arrays then `axes_order` must start with `C`")  # pragma: no cover
        pos: typing.List[typing.Union[slice, int]] = [slice(None) for _ in range(data.ndim)]
        c_pos = axes_order.index("C")
        res = []
        for i in range(data.shape[c_pos]):
            pos[c_pos] = i
            res.append(cls.reorder_axes(data[tuple(pos)], axes_order.replace("C", "")))
        return res

    @staticmethod
    def _merge_channel_names(
        base_channel_names: typing.List[str], new_channel_names: typing.List[str]
    ) -> typing.List[str]:
        base_channel_names = base_channel_names[:]
        reg = re.compile(r"channel \d+")
        for name in new_channel_names:
            match = reg.match(name)
            new_name = name
            base_name = name
            if match and base_name in base_channel_names:
                new_name = f"channel {len(base_channel_names) + 1}"
            i = 1
            while new_name in base_channel_names:
                new_name = f"{base_name} ({i})"
                i += 1
                if i > 10000:  # pragma: no cover
                    raise ValueError("fail when try to fix channel name")
            base_channel_names.append(new_name)
        return base_channel_names

    def merge(self, image: "Image", axis: str) -> "Image":
        """
        Produce new image merging image data along given axis. All metadata
        are obtained from self.

        :param Image image: Image to be merged
        :param str axis:
        :return: New image produced from merge
        :rtype: Image
        """
        if axis == "C":
            data = self._image_data_normalize(
                self._channel_arrays + [self.reorder_axes(x, image.array_axis_order) for x in image._channel_arrays]
            )
            channel_names = self._merge_channel_names(self.channel_names, image.channel_names)
        else:
            index = self.array_axis_order.index(axis)
            data = self._image_data_normalize(
                [
                    np.concatenate((y, self.reorder_axes(y, image.array_axis_order)), axis=index)
                    for x, y in zip(self._channel_arrays, image._channel_arrays)
                ]
            )
            channel_names = self.channel_names

        return self.substitute(data=data, ranges=self.ranges + image.ranges, channel_names=channel_names)

    @property
    def channel_names(self) -> typing.List[str]:
        return self._channel_names[:]

    @property
    def channel_pos(self) -> int:  # pragma: no cover
        """Channel axis. Need to have 'C' in :py:attr:`axis_order`"""
        warnings.warn(
            "channel_pos is deprecated and code its using may not work properly", category=FutureWarning, stacklevel=2
        )
        return self.axis_order.index("C")

    @property
    def x_pos(self):
        return self.array_axis_order.index("X")

    @property
    def y_pos(self):
        return self.array_axis_order.index("Y")

    @property
    def time_pos(self):
        """Time axis. Need to have 'T' in :py:attr:`axis_order`"""
        return self.array_axis_order.index("T")

    @property
    def stack_pos(self) -> int:
        """Stack axis. Need to have 'Z' in :py:attr:`axis_order`"""
        return self.array_axis_order.index("Z")

    @property
    def dtype(self) -> np.dtype:
        """dtype of image array"""
        return self._channel_arrays[0].dtype

    @staticmethod
    def _reorder_axes(array: np.ndarray, input_axes: str, return_axes) -> np.ndarray:
        if array.ndim != len(input_axes):
            raise ValueError(f"array.ndim ({array.ndim}) need to be equal to length of axes ('{input_axes}')")
        if input_axes == return_axes:
            return array
        mapping_dict = {v: i for i, v in enumerate(return_axes)}
        if array.ndim < len(return_axes):
            array = array.reshape(array.shape + (1,) * (len(return_axes) - array.ndim))
        new_positions = [mapping_dict[x] for x in input_axes if x in mapping_dict]
        axes_to_map = [i for i, x in enumerate(input_axes) if x in mapping_dict]
        return np.moveaxis(array, axes_to_map, new_positions)

    @classmethod
    def reorder_axes(cls, array: np.ndarray, axes: str) -> np.ndarray:
        """
        reorder axes to internal storage format

        :param np.ndarray array: array to have changed order of axes
        :param str axes: axes order
        :return: array with correct order of axes
        """
        return cls._reorder_axes(array, axes, cls.array_axis_order)

    def get_dimension_number(self) -> int:
        """return number of nontrivial dimensions"""
        return np.squeeze(self._channel_arrays[0]).ndim

    def get_dimension_letters(self) -> str:
        """
        :return: letters which indicates non trivial dimensions
        """
        return "".join(key for val, key in zip(self._channel_arrays[0].shape, self.array_axis_order) if val > 1)

    def substitute(
        self,
        data=None,
        image_spacing=None,
        file_path=None,
        mask=_DEF,
        default_coloring=None,
        ranges=None,
        channel_names=None,
    ) -> "Image":
        """Create copy of image with substitution of not None elements"""
        data = self._channel_arrays if data is None else data
        image_spacing = self._image_spacing if image_spacing is None else image_spacing
        file_path = self.file_path if file_path is None else file_path
        mask = self._mask_array if mask is _DEF else mask
        default_coloring = self.default_coloring if default_coloring is None else default_coloring
        ranges = self.ranges if ranges is None else ranges
        channel_names = self.channel_names if channel_names is None else channel_names
        return self.__class__(
            data=data,
            image_spacing=image_spacing,
            file_path=file_path,
            mask=mask,
            default_coloring=default_coloring,
            ranges=ranges,
            channel_names=channel_names,
            axes_order=self.axis_order,
        )

    def set_mask(self, mask: typing.Optional[np.ndarray], axes: typing.Optional[str] = None):
        """
        Set mask for image, check if it has proper shape.

        :param mask: mask in same shape like image. May not contains 1 dim axes.
        :param axes: order of axes in mask, use if different than :py:attr:`return_order`
        :raise ValueError: on wrong shape
        """
        if mask is None:
            self._mask_array = None
        elif axes is not None:
            self._mask_array = self.fit_mask_to_image(self.reorder_axes(mask, axes))
        else:
            self._mask_array = self.fit_mask_to_image(mask)

    def get_data(self) -> np.ndarray:
        if "C" in self.axis_order:
            return np.stack(self._channel_arrays, axis=self.axis_order.index("C"))
        return self._channel_arrays[0]

    @property
    def mask(self) -> typing.Optional[np.ndarray]:
        return self._mask_array[:] if self._mask_array is not None else None

    @staticmethod
    def _fit_array_to_image(base_shape, array: np.ndarray) -> np.ndarray:
        """change shape of array with inserting single dimensional entries"""
        shape = list(array.shape)
        for i, el in enumerate(base_shape):
            if el == 1 and el != shape[i]:
                shape.insert(i, 1)
            elif el != shape[i]:
                raise ValueError(f"Wrong array shape {shape} for {base_shape}")
        if len(shape) != len(base_shape):
            raise ValueError(f"Wrong array shape {shape} for {base_shape}")
        return np.reshape(array, shape)

    def fit_array_to_image(self, array: np.ndarray) -> np.ndarray:
        """
        Change shape of array with inserting single dimensional entries

        :param np.ndarray array: array to be fitted

        :return: reshaped array with added missing 1 in shape

        :raises ValueError: if cannot fit array
        """
        return self._fit_array_to_image(self._channel_arrays[0].shape, array)

    # noinspection DuplicatedCode
    def fit_mask_to_image(self, array: np.ndarray) -> np.ndarray:
        """
        call :py:meth:`fit_array_to_image` and
        then relabel and change type to minimal
        which fit all information
        """
        array = self.fit_array_to_image(array)
        if np.max(array) == 1:
            return array.astype(np.uint8)
        unique = np.unique(array)
        if unique.size == 1:
            if unique[0] != 0:
                return np.ones(array.shape, dtype=np.uint8)
            return array.astype(np.uint8)
        max_val = np.max(unique)
        return reduce_array(array, unique, max_val)

    def get_image_for_save(self) -> np.ndarray:
        """
        :return: numpy array in imagej tiff order axes
        """
        if "C" in self.axis_order:
            return self._reorder_axes(
                np.stack(self._channel_arrays, axis=self.axis_order.index("C")), self.axis_order, "TZCYX"
            )
        return self._reorder_axes(self._channel_arrays[0], self.axis_order, "TZCYX")

    def get_mask_for_save(self) -> typing.Optional[np.ndarray]:
        """
        :return: if image has mask then return mask with axes in proper order
        """
        if self._mask_array is None:
            return None
        return self._reorder_axes(self._mask_array, "".join(self.array_axis_order), "TZCYX")

    @property
    def has_mask(self) -> bool:
        """check if image is masked"""
        return self._mask_array is not None

    @property
    def is_time(self) -> bool:
        """check if image contains time data"""
        return self.times > 1

    @property
    def is_stack(self) -> bool:
        """check if image contain 3d data"""
        return self.layers > 1

    @property
    def channels(self) -> int:
        """number of image channels"""
        return len(self._channel_arrays)

    @property
    def layers(self) -> int:
        """z-dim of image"""
        return self._channel_arrays[0].shape[self.stack_pos]

    @property
    def times(self) -> int:
        """number of time frames"""
        return self._channel_arrays[0].shape[self.time_pos]

    @property
    def plane_shape(self) -> typing.Tuple[int, int]:
        """y,x size of image"""
        return self._channel_arrays[0].shape[self.y_pos], self._channel_arrays[0].shape[self.x_pos]

    @property
    def shape(self):
        """Whole image shape. order of axes my change. Current order is in :py:attr:`return_order`"""
        return self._channel_arrays[0].shape

    def swap_time_and_stack(self):
        """
        Swap time and stack axes.
        For example my be used to convert time image in 3d image.
        """
        image_array_list = [np.swapaxes(x, self.time_pos, self.stack_pos) for x in self._channel_arrays]
        return self.substitute(data=self._image_data_normalize(image_array_list))

    @classmethod
    def get_axis_positions(cls) -> typing.Dict[str, int]:
        """
        :return: dict with mapping axis to its position
        :rtype: dict
        """
        return {letter: i for i, letter in enumerate(cls.axis_order)}

    @classmethod
    def get_array_axis_positions(cls) -> typing.Dict[str, int]:
        """
        :return: dict with mapping axis to its position for array fitted to image
        :rtype: dict
        """
        return {letter: i for i, letter in enumerate(cls.array_axis_order)}

    def get_data_by_axis(self, **kwargs) -> np.ndarray:
        """
        Get part of data extracted by sub axis. Axis is selected by single letter from :py:attr:`axis_order`

        :param kwargs: axis list with
        :return:
        :rtype:
        """
        slices: typing.List[typing.Union[int, slice]] = [slice(None) for _ in range(len(self.array_axis_order))]
        axis_pos = self.get_array_axis_positions()
        if "c" in kwargs:
            kwargs["C"] = kwargs.pop("c")
        if "C" in kwargs and isinstance(kwargs["C"], str):
            kwargs["C"] = self.channel_names.index(kwargs["C"])

        channel = kwargs.pop("C", slice(None) if "C" in self.axis_order else 0)
        if isinstance(channel, Channel):
            channel = channel.value

        axis_order = self.axis_order
        for name, value in kwargs.items():
            if name.upper() in axis_pos:
                slices[axis_pos[name.upper()]] = value
                if isinstance(value, int):
                    axis_order = axis_order.replace(name.upper(), "")

        slices_t = tuple(slices)
        if isinstance(channel, int):
            return self._channel_arrays[channel][slices_t]
        return np.stack([x[slices_t] for x in self._channel_arrays[channel]], axis=axis_order.index("C"))

    def clip_array(self, array: np.ndarray, **kwargs: typing.Union[int, slice]) -> np.ndarray:
        """
        Clip array by axis. Axis is selected by single letter from :py:attr:`axis_order`

        :param array: array to clip
        :param kwargs: mapping from axis to position or slice on this axis
        :return: clipped array
        """
        array = self.fit_array_to_image(array)
        slices: typing.List[typing.Union[int, slice]] = [slice(None) for _ in range(len(self.array_axis_order))]
        axis_pos = self.get_array_axis_positions()
        for name in kwargs:
            if (n := name.upper()) in axis_pos:
                slices[axis_pos[n]] = kwargs[name]
        return array[tuple(slices)]

    def get_channel(self, num: Union[int, str, Channel]) -> np.ndarray:
        """
        Alias for :py:func:`get_sub_data` with argument ``c=num``

        :param int | str | Channel num: channel num or name to be extracted
        :return: given channel array
        :rtype: numpy.ndarray
        """
        return self.get_data_by_axis(c=num)

    def get_layer(self, time: int, stack: int) -> np.ndarray:
        """
        return single layer contains data for all channel

        :param time: time coordinate. For images with not time use 0.
        :param stack: "z coordinate. For time data use 0.
        :return:
        """
        warnings.warn(
            "Image.get_layer is deprecated. Use get_data_by_axis instead", category=DeprecationWarning, stacklevel=2
        )
        return self.get_data_by_axis(T=time, Z=stack)

    @property
    def is_2d(self) -> bool:
        """
        Check if image z and time dimension are equal to 1.
        Equivalent to:
        `image.layers == 1 and image.times == 1`
        """
        return self.layers == 1 and self.times == 1

    @property
    def spacing(self) -> Spacing:
        """image spacing"""
        return tuple(self._image_spacing[1:]) if self.is_2d else self._image_spacing

    def normalized_scaling(self, factor=DEFAULT_SCALE_FACTOR) -> Spacing:
        if self.is_2d:
            return (1, 1, *tuple(np.multiply(self.spacing, factor)))
        return (1, *tuple(np.multiply(self.spacing, factor)))

    @property
    def shift(self):
        return self._shift[1:] if self.is_2d else self._shift

    @property
    def voxel_size(self) -> Spacing:
        """alias for spacing"""
        return self.spacing

    def set_spacing(self, value: Spacing):
        """set image spacing"""
        if 0 in value:
            return
        if self.is_2d and len(value) + 1 == len(self._image_spacing):
            value = (1.0, *tuple(value))
        if len(value) != len(self._image_spacing):  # pragma: no cover
            raise ValueError("Correction of spacing fail.")
        self._image_spacing = tuple(value)

    @staticmethod
    def _frame_array(array: typing.Optional[np.ndarray], index_to_add: typing.List[int], frame=FRAME_THICKNESS):
        if array is None:  # pragma: no cover
            return array
        result_shape = list(array.shape)
        image_pos = [slice(None) for _ in range(array.ndim)]

        for index in index_to_add:
            result_shape[index] += frame * 2
            image_pos[index] = slice(frame, result_shape[index] - frame)

        data = np.zeros(shape=result_shape, dtype=array.dtype)
        data[tuple(image_pos)] = array
        return data

    @staticmethod
    def calc_index_to_frame(array_axis: str, important_axis: str) -> typing.List[int]:
        """
        calculate in which axis frame should be added

        :param str array_axis: list of image axis
        :param str important_axis: list of framed axis
        :return: list of indices to add frame.
        """
        return [array_axis.index(letter) for letter in important_axis]

    def _frame_cut_area(self, cut_area: typing.Iterable[slice], frame: int):
        cut_area = list(cut_area)
        important_axis = "XY" if self.is_2d else "XYZ"
        for ind in self.calc_index_to_frame(self.array_axis_order, important_axis):
            sl = cut_area[ind]
            cut_area[ind] = slice(
                max(sl.start - frame, 0) if sl.start is not None else None,
                sl.stop + frame if sl.stop is not None else None,
                sl.step,
            )
        return cut_area

    def _cut_image_slices(
        self, cut_area: typing.Iterable[slice], frame: int
    ) -> typing.Tuple[typing.List[np.ndarray], typing.Optional[np.ndarray]]:
        new_mask = None
        cut_area = self._frame_cut_area(cut_area, frame)
        new_image = [x[tuple(cut_area)] for x in self._channel_arrays]
        if self._mask_array is not None:
            new_mask = self._mask_array[tuple(cut_area)]
        return new_image, new_mask

    def _roi_to_slices(self, roi: np.ndarray) -> typing.List[slice]:
        cut_area = self.fit_array_to_image(roi)
        points = np.nonzero(cut_area)
        lower_bound = np.min(points, axis=1)
        upper_bound = np.max(points, axis=1)
        return [slice(x, y + 1) for x, y in zip(lower_bound, upper_bound)]

    def _cut_with_roi(self, cut_area: np.ndarray, replace_mask: bool, frame: int):
        new_mask = None
        cut_area = self.fit_array_to_image(cut_area)
        new_cut = tuple(self._roi_to_slices(cut_area))
        catted_cut_area = cut_area[new_cut]
        new_image = [np.copy(x[new_cut]) for x in self._channel_arrays]
        for el in new_image:
            el[catted_cut_area == 0] = 0
        if replace_mask:
            new_mask = catted_cut_area
        elif self._mask_array is not None:
            new_mask = self._mask_array[new_cut]
            new_mask[catted_cut_area == 0] = 0
        important_axis = "XY" if self.is_2d else "XYZ"
        new_image = [
            self._frame_array(x, self.calc_index_to_frame(self.array_axis_order, important_axis), frame)
            for x in new_image
        ]

        new_mask = self._frame_array(new_mask, self.calc_index_to_frame(self.array_axis_order, important_axis), frame)
        return new_image, new_mask

    def cut_image(
        self,
        cut_area: typing.Union[np.ndarray, typing.Iterable[slice]],
        replace_mask=False,
        frame: int = FRAME_THICKNESS,
        zero_out_cut_area: bool = True,
    ) -> "Image":
        """
        Create new image base on mask or list of slices
        :param bool replace_mask: if cut area is represented by mask array,
        then in result image the mask is set base on cut_area if cur_area is np.ndarray
        :param typing.Union[np.ndarray, typing.Iterable[slice]] cut_area: area to cut. Defined with slices or mask
        :param int frame: additional frame around cut_area
        :param bool zero_out_cut_area:
        :return: Image
        """
        if isinstance(cut_area, np.ndarray):
            if zero_out_cut_area:
                new_image, new_mask = self._cut_with_roi(cut_area, replace_mask, frame)
            else:
                new_cut = self._roi_to_slices(cut_area)
                new_image, new_mask = self._cut_image_slices(new_cut, frame)
                if replace_mask:
                    new_mask = cut_area[tuple(self._frame_cut_area(new_cut, frame))]
        else:
            new_image, new_mask = self._cut_image_slices(cut_area, frame)

        return self.__class__(
            data=self._image_data_normalize(new_image),
            image_spacing=self._image_spacing,
            file_path=None,
            mask=new_mask,
            default_coloring=self.default_coloring,
            ranges=self.ranges,
            channel_names=self.channel_names,
            axes_order=self.axis_order,
        )

    def get_imagej_colors(self):
        # TODO review
        if self.default_coloring is None:
            return None
        try:
            if len(self.default_coloring) != self.channels:
                return None
        except TypeError:
            return None
        res = []
        for color in self.default_coloring:
            if color.ndim == 1:
                res.append(np.array([np.linspace(0, x, num=256) for x in color]))
            else:
                if color.shape[1] != 256:
                    res.append(
                        np.array(
                            [
                                np.interp(np.linspace(0, 255, num=256), np.linspace(0, color.shape[1], num=256), x)
                                for x in color
                            ]
                        )
                    )
                res.append(color)
        return res

    def get_colors(self):
        # TODO review
        if self.default_coloring is None:
            return None
        res = []
        for color in self.default_coloring:
            if color.ndim == 2:
                res.append(list(color[:, -1]))
            else:
                res.append(list(color))
        return res

    def get_um_spacing(self) -> Spacing:
        """image spacing in micrometers"""
        return tuple(float(x * 10**6) for x in self.spacing)

    def get_um_shift(self) -> Spacing:
        """image spacing in micrometers"""
        return tuple(float(x * 10**6) for x in self.shift)

    def get_ranges(self) -> typing.List[typing.Tuple[float, float]]:
        """image brightness ranges for each channel"""
        return self.ranges[:]

    def __str__(self):
        return (
            f"{self.__class__} Shape {self._channel_arrays[0].shape}, dtype: {self._channel_arrays[0].dtype}, "
            f"labels: {self.channel_names}, coloring: {self.get_colors()} mask: {self.has_mask}"
        )

    def __repr__(self):
        mask_info = f"mask=True, mask_dtype={self._mask_array.dtype}" if self.mask is not None else "mask=False"
        return (
            f"Image(shape={self._channel_arrays[0].shape} dtype={self._channel_arrays[0].dtype}, spacing={self.spacing}"
            f", labels={self.channel_names}, channels={self.channels}, axes={self.axis_order!r}, {mask_info})"
        )

    @classmethod
    def _image_data_normalize(cls, data: _IMAGE_DATA) -> _IMAGE_DATA:
        if isinstance(data, np.ndarray):
            return data
        if cls.axis_order.startswith("C"):
            shape = data[0].shape
            if any(x.shape != shape for x in data):
                raise ValueError(f"Shape of arrays are different {[x.shape for x in data]}")
            return data
        if "C" not in cls.axis_order:
            return data[0]
        return np.stack(data, axis=cls.axis_order.index("C"))



import typing

from napari.utils.colormaps.colormap import Colormap
from napari.utils.colormaps.colormap_utils import AVAILABLE_COLORMAPS
from pydantic import Field

from PartSegCore.color_image.color_data import sitk_labels
from PartSegCore.utils import BaseModel

Num = typing.Union[int, float]


class Color(BaseModel):
    """
    store color information

    :param red: red color value
    :param green: green color value
    :param blue: blue color value
    :param alpha: alpha value
    """

    red: float = Field(ge=0.0, le=1.0)
    green: float = Field(ge=0.0, le=1.0)
    blue: float = Field(ge=0.0, le=1.0)
    alpha: float = Field(1, ge=0.0, le=1.0)

    def as_tuple(self):
        return (self.red, self.green, self.blue, self.alpha)

    @classmethod
    def from_tuple(cls, tup):
        """
        create color from tuple

        :param tup: tuple of color values
        :return: color
        """
        return cls(red=tup[0], green=tup[1], blue=tup[2], alpha=tup[3] if len(tup) > 3 else 1)

    def is_close(self, other: "Color", epsilon: float = 1e-6) -> bool:
        """
        check if color is close to other color

        :param other: other color
        :param epsilon: epsilon value
        :return: True if colors are close
        """
        return (
            abs(self.red - other.red) < epsilon
            and abs(self.green - other.green) < epsilon
            and abs(self.blue - other.blue) < epsilon
            and abs(self.alpha - other.alpha) < epsilon
        )


starting_colors = ["red", "green", "blue", "magenta", "inferno", "magma"]

default_colormap_dict = {name: AVAILABLE_COLORMAPS[name] for name in starting_colors}
default_colormap_dict.update(AVAILABLE_COLORMAPS)
default_colormap_dict.update(
    {
        f"{k}_reversed": Colormap(v.colors[::-1], controls=1 - v.controls[::-1])
        for k, v in AVAILABLE_COLORMAPS.items()
        if not k.endswith("_k")
    }
)


default_label_dict = {"default": sitk_labels}



import typing

import numpy as np
import SimpleITK as sitk
from local_migrator import register_class

from PartSegCore.image_operations import RadiusType, dilate, erode
from PartSegCore.utils import BaseModel
from PartSegImage.image import minimal_dtype


@register_class(old_paths=["PartSeg.utils.mask_create.MaskProperty"])
class MaskProperty(BaseModel):
    """
    Description of creation mask from segmentation

    :ivar RadiusType ~.dilate: Select dilation mode.
    :ivar int ~.dilate_radius: Radius of dilation calculate with respect of image spacing.
    :ivar RadiusType ~.fill_holes: Select if fill holes and if it should be done in 2d or 3d.
    :ivar int ~.max_holes_size: Maximum holes size if positive. Otherwise fill all holes.
    :ivar bool ~.save_components: If mask should save components of segmentation or set to 1.
    :ivar bool ~.clip_to_mask: If resulted should be clipped to previous mask (if exist).
        Useful for positive dilate radius
    :ivar bool ~.reversed_mask: If mask should be reversed (region which are not part of segmentation)
    """

    __readonly__ = False

    dilate: RadiusType
    dilate_radius: int
    fill_holes: RadiusType
    max_holes_size: int
    save_components: bool
    clip_to_mask: bool
    reversed_mask: bool = False

    def __str__(self):
        return (
            f"Mask property\ndilate: {self.dilate}\n"
            + (f"dilate radius {self.dilate_radius}\n" if self.dilate != RadiusType.NO else "")
            + f"fill holes: {self.fill_holes}\n"
            + (f"max holes size: {self.max_holes_size}\n" if self.fill_holes != RadiusType.NO else "")
            + f"save components: {self.save_components}\nclip to mask: {self.clip_to_mask}\n"
            + f"reversed mask {self.reversed_mask}"
        )

    def __repr__(self):
        return (
            f"Mask MaskProperty(dilate={self.dilate}, dilate radius={self.dilate_radius}, "
            f"fill holes={self.fill_holes}, max holes size={self.max_holes_size}, "
            f"save components={self.save_components}, clip to mask={self.clip_to_mask}, "
            f"reversed mask={self.reversed_mask})"
        )

    def __eq__(self, other):
        return (
            isinstance(other, MaskProperty)
            and self.dilate == other.dilate
            and self.dilate_radius == other.dilate_radius
            and self.fill_holes == other.fill_holes
            and self.max_holes_size == other.max_holes_size
            and self.clip_to_mask == other.clip_mask
            and self.reversed_mask == other.reverse_mask
        )

    @classmethod
    def simple_mask(cls) -> "MaskProperty":
        """
        Create MaskProperty which describe simple conversion segmentation to mask

        :rtype: MaskProperty
        """
        return cls(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.NO,
            max_holes_size=0,
            save_components=False,
            clip_to_mask=False,
        )


def mp_eq(self: MaskProperty, other: MaskProperty):
    """Compare two :class:`MaskProperty`"""
    return (
        self.__class__ == other.__class__
        and self.dilate == other.dilate
        and self.fill_holes == other.fill_holes
        and self.save_components == other.save_components
        and self.clip_to_mask == other.clip_to_mask
        and (self.dilate == RadiusType.NO or (self.dilate_radius == other.dilate_radius))
        and (self.fill_holes == RadiusType.NO or (self.max_holes_size == other.max_holes_size))
        and (self.reversed_mask == other.reversed_mask)
    )


MaskProperty.__eq__ = mp_eq


def calculate_mask(
    mask_description: MaskProperty,
    roi: np.ndarray,
    old_mask: typing.Optional[np.ndarray],
    spacing: typing.Iterable[typing.Union[float, int]],
    components: typing.Optional[typing.List[int]] = None,
    time_axis: typing.Optional[int] = 0,
) -> np.ndarray:
    """
    Function for calculate mask base on MaskProperty.
    If dilate_radius is negative then holes closing is done before erode,
    otherwise it is done after dilate

    :param MaskProperty mask_description: information how calculate mask
    :param np.ndarray roi: array on which mask is calculated
    :param typing.Optional[np.ndarray] old_mask: if in mask_description there is set to crop and old_mask is not None
        then final mask is clipped to this area
    :param typing.Iterable[typing.Union[float,int]] spacing: spacing of image. Needed for calculating radius of dilate
    :param typing.Optional[typing.List[int]] components: If present inform which components
        should be used when calculation mask, otherwise use all.
    :param typing.Optional[int] time_axis: which axis of array should be treated as time. IF none then none.
    :return: new mask
    :rtype: np.ndarray
    """
    spacing_min = min(spacing)
    spacing = [x / spacing_min for x in spacing]
    dilate_radius = [int(abs(mask_description.dilate_radius / x) + 0.5) for x in spacing]
    if mask_description.dilate == RadiusType.R2D:
        dilate_radius = dilate_radius[-2:]
    if components is not None:
        components_num = max(np.max(roi), *components) + 1
        map_array = np.zeros(components_num, dtype=minimal_dtype(components_num))
        for el in components:
            map_array[el] = el
        roi = map_array[roi]

    mask = np.copy(roi) if mask_description.save_components else np.array(roi > 0)
    if time_axis is None:
        return _calculate_mask(mask_description, dilate_radius, mask, old_mask)
    slices: typing.List[typing.Union[slice, int]] = [slice(None) for _ in range(mask.ndim)]
    final_shape = list(mask.shape)
    final_shape[time_axis] = 1
    final_shape = tuple(final_shape)
    res = []
    for i in range(mask.shape[time_axis]):
        slices[time_axis] = i
        t_slices = tuple(slices)
        _old_mask = old_mask[t_slices] if old_mask is not None else None
        res.append(_calculate_mask(mask_description, dilate_radius, mask[t_slices], _old_mask).reshape(final_shape))
    return np.concatenate(res, axis=time_axis)


def _calculate_mask(
    mask_description: MaskProperty,
    dilate_radius: typing.List[int],
    mask: np.ndarray,
    old_mask: typing.Union[None, np.ndarray],
) -> np.ndarray:
    if mask_description.dilate != RadiusType.NO and mask_description.dilate_radius != 0:
        if mask_description.dilate_radius > 0:
            mask = dilate(mask, dilate_radius, mask_description.dilate == RadiusType.R2D)
            mask = _fill_holes(mask_description, mask)
        elif mask_description.dilate_radius < 0:
            mask = _fill_holes(mask_description, mask)
            mask = erode(mask, dilate_radius, mask_description.dilate == RadiusType.R2D)
    elif mask_description.fill_holes != RadiusType.NO:
        mask = _fill_holes(mask_description, mask)
    if mask_description.reversed_mask:
        mask = np.array(mask == 0).astype(np.uint8)
    if mask_description.clip_to_mask and old_mask is not None:
        mask[old_mask == 0] = 0
    return mask


def _cut_components(
    mask: np.ndarray, image: np.ndarray, borders: int = 0
) -> typing.Iterator[typing.Tuple[np.ndarray, typing.List[slice], int]]:
    sizes = np.bincount(mask.flat)
    for i, size in enumerate(sizes[1:], 1):
        if size > 0:
            points = np.nonzero(mask == i)
            lower_bound = np.min(points, axis=1)
            upper_bound = np.max(points, axis=1)
            new_cut = tuple(slice(x, y + 1) for x, y in zip(lower_bound, upper_bound))
            new_size = [y - x + 1 + 2 * borders for x, y in zip(lower_bound, upper_bound)]
            if borders > 0:
                res = np.zeros(new_size, dtype=image.dtype)
                res_cut = tuple(slice(borders, x - borders) for x in tuple(res.shape))
                tmp_res = np.copy(image[new_cut])
                tmp_res[mask[new_cut] != i] = 0
                res[res_cut] = tmp_res
            else:
                res = image[new_cut]
                res[mask[new_cut] != i] = 0
            yield res, tuple(new_cut), i


def _fill_holes(mask_description: MaskProperty, mask: np.ndarray) -> np.ndarray:
    if mask_description.fill_holes == RadiusType.NO:
        return mask
    if mask_description.save_components:
        border = 1
        res_slice = tuple(slice(border, -border) for _ in range(mask.ndim))
        mask_description_copy = mask_description.copy(update={"save_components": False})
        mask_prohibited = mask > 0
        for component, slice_arr, cmp_num in _cut_components(mask, mask, border):
            mask_prohibited_component = mask_prohibited[slice_arr]
            new_component = _fill_holes(mask_description_copy, component)
            new_component = new_component[res_slice]
            new_component[mask_prohibited_component > 0] = 0
            mask[slice_arr][new_component > 0] = cmp_num
    else:
        if mask_description.fill_holes == RadiusType.R2D:
            mask = fill_2d_holes_in_mask(mask, mask_description.max_holes_size)
        elif mask_description.fill_holes == RadiusType.R3D:
            mask = fill_holes_in_mask(mask, mask_description.max_holes_size)
        mask = mask.astype(bool)
    return mask


def fill_holes_in_mask(mask: np.ndarray, volume: int) -> np.ndarray:
    """
    Fil holes in mask. If volume has positive Value then fill holes only smaller than this value

    :param mask: mask to be modified
    :param volume: maximum volume of holes which will be filled
    :return: modified mask
    """
    holes_mask = (mask == 0).astype(np.uint8)
    component_mask = sitk.GetArrayFromImage(
        sitk.RelabelComponent(sitk.ConnectedComponent(sitk.GetImageFromArray(holes_mask)))
    )
    border_set: typing.Set[int] = set()
    for dim_num in range(component_mask.ndim):
        border_set.update(list(np.unique(np.take(component_mask, [0, -1], axis=dim_num))))
    if 0 in border_set:
        border_set.remove(0)
    components_num = component_mask.max()
    if component_mask.dtype.type(components_num) >= component_mask.dtype.type(components_num + 1):
        raise ValueError("overflow error")
    for num in border_set:
        component_mask[component_mask == num] = components_num + 1
    if volume > 0:
        sizes = np.bincount(component_mask.flat)
        for i, v in enumerate(sizes[1:], 1):
            if v <= volume and i < components_num + 1:
                component_mask[component_mask == i] = 0
    else:
        component_mask[component_mask <= components_num] = 0
    return component_mask == 0


def fill_2d_holes_in_mask(mask: np.ndarray, volume: int) -> np.ndarray:
    """
    fill holes in each 2d layer separately

    :param mask: mask to fill holes
    :param volume: minimum volume
    :return: modified mask
    """
    mask = np.copy(mask)
    if mask.ndim == 2:
        return fill_holes_in_mask(mask, volume)
    for i in range(mask.shape[0]):
        mask[i] = fill_holes_in_mask(mask[i], volume)
    return mask



import json
import os.path
import tarfile
import typing
from io import BytesIO
from pathlib import Path

import h5py
import numpy as np
import tifffile

from PartSegCore.algorithm_describe_base import AlgorithmProperty, Register
from PartSegCore.analysis.io_utils import ProjectTuple, project_version_info
from PartSegCore.io_utils import NotSupportedImage, SaveBase, SaveMaskAsTiff, SaveROIAsNumpy, SaveROIAsTIFF, get_tarinfo
from PartSegCore.json_hooks import PartSegEncoder
from PartSegCore.project_info import HistoryElement
from PartSegCore.roi_info import ROIInfo
from PartSegCore.universal_const import UNIT_SCALE, Units
from PartSegImage import Channel, Image, ImageWriter

__all__ = [
    "SaveProject",
    "SaveCmap",
    "SaveXYZ",
    "SaveAsTiff",
    "SaveAsNumpy",
    "SaveProfilesToJSON",
    "save_dict",
]


# TODO add progress function to io
def save_project(
    file_path: str,
    image: Image,
    roi_info: ROIInfo,
    mask: typing.Optional[np.ndarray],
    history: typing.List[HistoryElement],
    algorithm_parameters: dict,
):
    # TODO add support for binary objects
    ext = os.path.splitext(file_path)[1]
    tar_mod = "w:bz2" if ext.lower() in [".bz2", ".tbz2"] else "w:gz"
    with tarfile.open(file_path, tar_mod) as tar:
        segmentation_buff = BytesIO()
        # noinspection PyTypeChecker
        tifffile.imwrite(segmentation_buff, roi_info.roi)
        segmentation_tar = get_tarinfo("segmentation.tif", segmentation_buff)
        tar.addfile(segmentation_tar, fileobj=segmentation_buff)
        if roi_info.alternative:
            alternative_buff = BytesIO()
            np.savez(alternative_buff, **roi_info.alternative)
            alternative_tar = get_tarinfo("alternative.npz", alternative_buff)
            tar.addfile(alternative_tar, fileobj=alternative_buff)
        if mask is not None:
            if mask.dtype == bool:
                mask = mask.astype(np.uint8)
            segmentation_buff = BytesIO()
            # noinspection PyTypeChecker
            tifffile.imwrite(segmentation_buff, mask)
            segmentation_tar = get_tarinfo("mask.tif", segmentation_buff)
            tar.addfile(segmentation_tar, fileobj=segmentation_buff)
        image_buff = BytesIO()
        ImageWriter.save(image, image_buff, compression=None)
        tar_image = get_tarinfo("image.tif", image_buff)
        tar.addfile(tarinfo=tar_image, fileobj=image_buff)
        para_str = json.dumps(algorithm_parameters, cls=PartSegEncoder)
        parameters_buff = BytesIO(para_str.encode("utf-8"))
        tar_algorithm = get_tarinfo("algorithm.json", parameters_buff)
        tar.addfile(tar_algorithm, parameters_buff)
        meta_str = json.dumps(
            {"project_version_info": str(project_version_info), "roi_annotations": roi_info.annotations},
            cls=PartSegEncoder,
        )
        meta_buff = BytesIO(meta_str.encode("utf-8"))
        tar_meta = get_tarinfo("metadata.json", meta_buff)
        tar.addfile(tar_meta, meta_buff)
        el_info = []
        for i, el in enumerate(history):
            el_info.append(
                {
                    "index": i,
                    "algorithm_name": el.roi_extraction_parameters["algorithm_name"],
                    "values": el.roi_extraction_parameters["values"],
                    "mask_property": el.mask_property,
                    "annotations": el.annotations,
                }
            )
            el.arrays.seek(0)
            hist_info = get_tarinfo(f"history/arrays_{i}.npz", el.arrays)
            el.arrays.seek(0)
            tar.addfile(hist_info, el.arrays)
        if el_info:
            hist_str = json.dumps(el_info, cls=PartSegEncoder)
            hist_buff = BytesIO(hist_str.encode("utf-8"))
            tar_algorithm = get_tarinfo("history/history.json", hist_buff)
            tar.addfile(tar_algorithm, hist_buff)


def _save_cmap(
    cmap_file: h5py.File,
    data: np.ndarray,
    spacing,
    segmentation: np.ndarray,
    reverse_base: float,
    cmap_profile: dict,
    metadata: typing.Optional[dict] = None,
):
    if cmap_profile["reverse"]:
        data = reverse_base - data
        data[data < 0] = 0
    data = np.copy(data)
    data[segmentation == 0] = 0
    grp = cmap_file.create_group("Chimera/image1")
    z, y, x = data.shape
    data_set = grp.create_dataset("data_zyx", (z, y, x), dtype="f", compression="gzip")
    data_set[...] = data

    if metadata:
        meta_group = cmap_file.create_group("Chimera/image1/Statistics")
        for key, val in metadata.items():
            meta_group.attrs[key] = val

    grp = cmap_file["Chimera"]
    grp.attrs["CLASS"] = np.bytes_("GROUP")
    grp.attrs["TITLE"] = np.bytes_("")
    grp.attrs["VERSION"] = np.bytes_("1.0")

    grp = cmap_file["Chimera/image1"]
    grp.attrs["CLASS"] = np.bytes_("GROUP")
    grp.attrs["TITLE"] = np.bytes_("")
    grp.attrs["VERSION"] = np.bytes_("1.0")
    grp.attrs["step"] = np.array(spacing, dtype=np.float32)[::-1] * UNIT_SCALE[cmap_profile["units"].value]

    return cmap_file


def save_cmap(
    file: typing.Union[str, h5py.File, BytesIO],
    data: np.ndarray,
    spacing,
    segmentation: np.ndarray,
    reverse_base: float,
    cmap_profile: dict,
    metadata: typing.Optional[dict] = None,
):
    if segmentation is None or np.max(segmentation) == 0:
        raise ValueError("No segmentation")
    if isinstance(file, (str, BytesIO, Path)):
        if isinstance(file, str) and os.path.exists(file):
            os.remove(file)
        with h5py.File(file, "w") as cmap_file:
            _save_cmap(cmap_file, data, spacing, segmentation, reverse_base, cmap_profile, metadata)
    elif isinstance(file, h5py.File):
        _save_cmap(file, data, spacing, segmentation, reverse_base, cmap_profile, metadata)
    else:
        raise ValueError(f"Wrong type of file argument, type: {type(file)}")


class SaveProject(SaveBase):
    @classmethod
    def get_name(cls):
        return "Project (*.tgz *.tbz2 *.gz *.bz2)"

    @classmethod
    def get_short_name(cls):
        return "project"

    @classmethod
    def get_fields(cls):
        return []

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info: ProjectTuple,
        parameters: typing.Optional[dict] = None,
        range_changed=None,
        step_changed=None,
    ):
        save_project(
            save_location,
            project_info.image,
            project_info.roi_info,
            project_info.mask,
            project_info.history,
            project_info.algorithm_parameters,
        )


class SaveCmap(SaveBase):
    @classmethod
    def get_name(cls):
        return "Chimera CMAP (*.cmap)"

    @classmethod
    def get_short_name(cls):
        return "cmap"

    @classmethod
    def get_fields(cls):
        return [
            AlgorithmProperty("channel", "Channel", 0, value_type=Channel),
            AlgorithmProperty("separated_objects", "Separate Objects", False),
            AlgorithmProperty("clip", "Clip area", False),
            AlgorithmProperty("units", "Units", Units.nm, value_type=Units),
            AlgorithmProperty(
                "reverse", "Reverse", False, help_text="Reverse brightness off image (for electron microscopy)"
            ),
        ]

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info: ProjectTuple,
        parameters: dict,
        range_changed=None,
        step_changed=None,
    ):
        if project_info.image.shape[project_info.image.time_pos] != 1:
            raise NotSupportedImage("This save method o not support time data")
        data = project_info.image.get_data_by_axis(c=parameters["channel"], t=0)
        spacing = project_info.image.spacing
        segmentation = project_info.image.clip_array(project_info.roi_info.roi, t=0)

        reverse_base = float(np.mean(data[segmentation == 0]))
        if parameters.get("clip", False):
            positions = np.transpose(np.nonzero(segmentation))
            clip_down = np.min(positions, 0)
            clip_up = np.max(positions, 0)
            clip = tuple(slice(x, y + 1) for x, y in zip(clip_down, clip_up))
            data = data[clip]
            segmentation = segmentation[clip]

        if parameters["separated_objects"] and isinstance(save_location, (str, Path)):
            for i in range(1, np.max(segmentation) + 1):
                seg = (segmentation == i).astype(np.uint8)
                if np.any(seg):
                    base, ext = os.path.splitext(save_location)
                    save_loc = f"{base}_comp{i}{ext}"
                    save_cmap(save_loc, data, spacing, seg, reverse_base, parameters)
        else:
            save_cmap(save_location, data, spacing, segmentation, reverse_base, parameters)


class SaveXYZ(SaveBase):
    @classmethod
    def get_name(cls):
        return "XYZ text (*.xyz *.txt)"

    @classmethod
    def get_short_name(cls):
        return "xyz"

    @classmethod
    def get_fields(cls):
        return [
            AlgorithmProperty("channel", "Channel", 0, value_type=Channel),
            AlgorithmProperty("separated_objects", "Separate Objects", False),
            AlgorithmProperty("clip", "Clip area", False),
        ]

    @classmethod
    def _save(cls, save_location, channel_image, segmentation_mask, shift):
        positions = np.transpose(np.nonzero(segmentation_mask))
        positions = np.flip(positions, 1)
        positions -= shift
        values = channel_image[segmentation_mask]
        values = values.reshape(values.size, 1)
        data = np.append(positions, values, axis=1)
        fm = "%d" if np.issubdtype(channel_image.dtype, np.integer) else "%f"
        # noinspection PyTypeChecker
        np.savetxt(save_location, data, fmt=["%d"] * channel_image.ndim + [fm], delimiter=" ")

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info: ProjectTuple,
        parameters: dict,
        range_changed=None,
        step_changed=None,
    ):
        if project_info.roi_info.roi is None:
            raise ValueError("Not segmentation")
        if isinstance(save_location, (str, Path)) and not os.path.exists(os.path.dirname(save_location)):
            os.makedirs(os.path.dirname(save_location))
        if parameters.get("separated_objects", False) and not isinstance(save_location, (str, Path)):
            raise ValueError("Saving components to buffer not supported")
        if project_info.image.shape[project_info.image.time_pos] != 1 and "time" not in parameters:
            raise NotSupportedImage("This save method o not support time data")
        channel_image = project_info.image.get_data_by_axis(c=parameters["channel"], t=parameters.get("time", 0))

        segmentation_mask = np.array(project_info.roi_info.roi > 0)
        segmentation_mask = project_info.image.clip_array(segmentation_mask, t=parameters.get("time", 0))
        if parameters.get("clip", False):
            positions = np.transpose(np.nonzero(segmentation_mask))
            positions = np.flip(positions, 1)
            shift = np.min(positions, 0)
        else:
            shift = np.array([0] * segmentation_mask.ndim)
        cls._save(save_location, channel_image, segmentation_mask, shift)
        if parameters.get("separated_objects", False):
            components_count = np.bincount(project_info.roi_info.roi.flat)
            for i, size in enumerate(components_count[1:], 1):
                if size > 0:
                    segmentation_mask = np.array(project_info.roi_info.roi == i)[parameters.get("time", 0)]
                    base_path, ext = os.path.splitext(save_location)
                    new_save_location = f"{base_path}_part{i}{ext}"
                    cls._save(new_save_location, channel_image, segmentation_mask, shift)


class SaveAsTiff(SaveBase):
    @classmethod
    def get_name(cls):
        return "Image (*.tiff *.tif)"

    @classmethod
    def get_short_name(cls):
        return "tiff"

    @classmethod
    def get_fields(cls):
        return []

    @classmethod
    def need_segmentation(cls):
        return False

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info,
        parameters: typing.Optional[dict] = None,
        range_changed=None,
        step_changed=None,
    ):
        ImageWriter.save(project_info.image, save_location)


class SaveAsNumpy(SaveBase):
    @classmethod
    def get_name(cls):
        return "Numpy Image (*.npy)"

    @classmethod
    def get_short_name(cls):
        return "numpy_image"

    @classmethod
    def get_fields(cls):
        return [
            AlgorithmProperty(
                "squeeze",
                "Squeeze  array",
                False,
                help_text="Remove single-dimensional entries from the shape of an array",
            )
        ]

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info,
        parameters: dict,
        range_changed=None,
        step_changed=None,
    ):
        data = project_info.image.get_data()
        if parameters["squeeze"]:
            data = np.squeeze(data)
        np.save(save_location, data)


class SaveProfilesToJSON(SaveBase):
    @classmethod
    def get_short_name(cls):
        return "json"

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info,
        parameters: dict,
        range_changed=None,
        step_changed=None,
    ):
        with open(save_location, "w", encoding="utf-8") as ff:
            json.dump(project_info, ff, cls=PartSegEncoder, indent=2)

    @classmethod
    def get_name(cls) -> str:
        return "Segment profile (*.json)"

    @classmethod
    def get_fields(cls) -> typing.List[typing.Union[AlgorithmProperty, str]]:
        return []


class SavePipelineToJSON(SaveProject):
    @classmethod
    def get_name(cls) -> str:
        return "Segment pipeline (*.json)"


save_dict = Register(
    SaveProject,
    SaveCmap,
    SaveXYZ,
    SaveAsTiff,
    SaveMaskAsTiff,
    SaveAsNumpy,
    SaveROIAsTIFF,
    SaveROIAsNumpy,
    class_methods=SaveBase.need_functions,
)



import importlib
import itertools
import pkgutil
import sys
import typing
from importlib.metadata import entry_points


def iter_entrypoints(group: str):
    if sys.version_info < (3, 10):
        return itertools.chain(
            entry_points().get(group, []),
            entry_points().get(group.lower(), []),
        )

    return itertools.chain(
        entry_points().select(group=group),
        entry_points().select(group=group.lower()),
    )


def get_plugins():
    packages = pkgutil.iter_modules(__path__, f"{__name__}.")
    packages2 = iter_entrypoints("PartSegCore.plugins")
    return [importlib.import_module(el.name) for el in packages] + [el.load() for el in packages2]


plugins_loaded = set()


def register():
    for el in get_plugins():
        if hasattr(el, "register") and el.__name__ not in plugins_loaded:
            if not isinstance(el.register, typing.Callable):  # pragma: no cover
                raise TypeError(f"Plugin {el.__name__} has no register method")
            el.register()
            plugins_loaded.add(el.__name__)


def register_if_need():
    if len(plugins_loaded) == 0:
        register()



from unittest.mock import MagicMock, Mock, patch

import numpy as np
import pytest
from qtpy.QtCore import Qt

from PartSeg._roi_analysis.main_window import ChannelProperty, MainWindow, Options
from PartSegCore.analysis import ProjectTuple
from PartSegCore.roi_info import ROIInfo
from PartSegCore.segmentation import ROIExtractionResult


class TestAnalysisMainWindow:
    # @pytest.mark.skipif((platform.system() == "Linux") and CI_BUILD, reason="debug test fail")
    @pytest.mark.pyside_skip()
    def test_opening(self, qtbot, tmpdir):
        main_window = MainWindow(tmpdir, initial_image=False)
        qtbot.addWidget(main_window)
        main_window.main_menu.batch_processing_btn.click()
        main_window.main_menu.advanced_btn.click()
        main_window.advanced_window.close()
        main_window.advanced_window.close()
        qtbot.wait(50)

    @pytest.mark.pyside_skip()
    def test_change_theme(self, qtbot, tmpdir):
        main_window = MainWindow(tmpdir, initial_image=False)
        qtbot.addWidget(main_window)
        assert main_window.raw_image.viewer.theme == "light"
        main_window.settings.theme_name = "dark"
        assert main_window.raw_image.viewer.theme == "dark"

    @pytest.mark.pyside_skip()
    def test_scale_bar(self, qtbot, tmpdir):
        main_window = MainWindow(tmpdir, initial_image=False)
        qtbot.addWidget(main_window)
        main_window._scale_bar_warning = False
        assert not main_window.result_image.viewer.scale_bar.visible
        main_window._toggle_scale_bar()
        assert main_window.result_image.viewer.scale_bar.visible

    def test_get_project_info(self, image, tmp_path):
        res = MainWindow.get_project_info(str(tmp_path / "test.tiff"), image)
        assert isinstance(res.roi_info, ROIInfo)
        assert isinstance(res, ProjectTuple)

        roi = np.zeros(image.shape, dtype=np.uint8)
        roi[:, 2:-2] = 1
        res = MainWindow.get_project_info(str(tmp_path / "test.tiff"), image, ROIInfo(roi))
        assert isinstance(res.roi_info, ROIInfo)
        assert set(res.roi_info.bound_info) == {1}


@pytest.fixture()
def analysis_options(qtbot, part_settings):
    ch_property = ChannelProperty(part_settings, "test")
    qtbot.addWidget(ch_property)
    left_image = MagicMock()
    synchronize = MagicMock()
    options = Options(part_settings, ch_property, left_image, synchronize)
    qtbot.addWidget(options)
    qtbot.addWidget(options.compare_btn)
    return options


class TestAnalysisOptions:
    def test_create(self, qtbot, analysis_options):
        assert analysis_options.choose_profile.count() == 1
        assert analysis_options.choose_profile.currentText() == "<none>"
        assert analysis_options.choose_pipe.count() == 1
        assert analysis_options.choose_pipe.currentText() == "<none>"

    def test_add_profile(self, qtbot, part_settings, analysis_options, lower_threshold_profile, border_rim_profile):
        part_settings.roi_profiles[lower_threshold_profile.name] = lower_threshold_profile
        assert analysis_options.choose_profile.count() == 2
        assert analysis_options.choose_profile.currentText() == "<none>"
        assert analysis_options.choose_profile.itemText(1) == lower_threshold_profile.name
        assert analysis_options.choose_profile.itemData(1, Qt.ItemDataRole.ToolTipRole) == str(lower_threshold_profile)
        part_settings.roi_profiles[border_rim_profile.name] = border_rim_profile
        assert analysis_options.choose_profile.count() == 3
        assert analysis_options.choose_pipe.count() == 1
        del part_settings.roi_profiles[lower_threshold_profile.name]
        assert analysis_options.choose_profile.count() == 2

    def test_add_pipeline(self, part_settings, analysis_options, sample_pipeline, sample_pipeline2):
        part_settings.roi_pipelines[sample_pipeline.name] = sample_pipeline
        assert analysis_options.choose_pipe.count() == 2
        assert analysis_options.choose_pipe.itemText(1) == sample_pipeline.name
        assert analysis_options.choose_pipe.itemData(1, Qt.ItemDataRole.ToolTipRole) == str(sample_pipeline)
        part_settings.roi_pipelines[sample_pipeline2.name] = sample_pipeline2
        assert analysis_options.choose_pipe.count() == 3
        assert analysis_options.choose_profile.count() == 1

    def test_compare_action(self, part_settings, analysis_options, qtbot):
        roi = np.zeros(part_settings.image.shape, dtype=np.uint8)
        roi[0, 2:10] = 1
        roi[0, 10:-2] = 2
        part_settings.roi = roi
        assert analysis_options.compare_btn.text() == "Compare"
        analysis_options.compare_action()
        assert analysis_options.compare_btn.text() == "Remove"
        assert part_settings.compare_segmentation is not None
        analysis_options.compare_action()
        assert analysis_options.compare_btn.text() == "Compare"
        assert part_settings.compare_segmentation is not None

    @patch("PartSeg._roi_analysis.main_window.QMessageBox.information")
    def test_empty_save_pipeline(self, info, analysis_options, part_settings):
        assert part_settings.history_size() == 0
        analysis_options.save_pipeline()
        info.assert_called_once()
        assert info.call_args[0][1] == "No mask created"

    @patch("PartSeg._roi_analysis.main_window.QMessageBox.information")
    def test_save_pipeline_no_segmentation(self, info, analysis_options, part_settings, history_element):
        part_settings.add_history_element(history_element)
        analysis_options.save_pipeline()
        info.assert_called_once()
        assert info.call_args[0][1] == "No segmentation"

    @patch("PartSeg._roi_analysis.main_window.QMessageBox.information")
    def test_save_pipeline_no_algorithm_values(
        self, info, analysis_options, part_settings, history_element, lower_threshold_profile
    ):
        part_settings.add_history_element(history_element)
        part_settings.last_executed_algorithm = lower_threshold_profile.name
        analysis_options.save_pipeline()
        info.assert_called_once()
        assert info.call_args[0][1] == "Some problem"

    @patch("PartSeg._roi_analysis.main_window.QInputDialog.getText", return_value=("test", True))
    def test_save_pipeline(self, info, analysis_options, part_settings, history_element, lower_threshold_profile):
        part_settings.add_history_element(history_element)
        part_settings.set_algorithm(f"algorithms.{lower_threshold_profile.name}", lower_threshold_profile)
        part_settings.last_executed_algorithm = lower_threshold_profile.name
        assert analysis_options.choose_pipe.count() == 1
        analysis_options.save_pipeline()
        info.assert_called_once()
        assert "test" in part_settings.roi_pipelines
        assert info.call_args[0][1] == "Pipeline name"
        assert analysis_options.choose_pipe.count() == 2
        assert analysis_options.choose_pipe.itemText(1) == "test"

    @patch("PartSeg._roi_analysis.main_window.QInputDialog.getText", return_value=("profile", True))
    def test_save_profile(self, info, analysis_options, part_settings):
        assert analysis_options.choose_profile.count() == 1
        analysis_options.save_profile()
        assert analysis_options.choose_profile.count() == 2
        assert analysis_options.choose_profile.itemText(1) == "profile"
        info.assert_called_once()

    def test_execution_done(self, analysis_options, part_settings, monkeypatch):
        sender_mock = MagicMock()
        sender_mock.get_info_text.return_value = "test"
        mock = Mock(return_value=sender_mock)
        info_mock = Mock()

        monkeypatch.setattr(analysis_options, "sender", mock)
        monkeypatch.setattr("qtpy.QtWidgets.QMessageBox.information", info_mock)

        res = ROIExtractionResult(
            roi=np.zeros(part_settings.image.shape, dtype="uint8"),
            parameters=analysis_options.algorithm_choose_widget.algorithm_dict[
                "Lower threshold"
            ].get_segmentation_profile(),
            info_text="info",
        )
        analysis_options.execution_done(res)

        assert analysis_options.label.toPlainText() == "test"
        info_mock.assert_called_once_with(analysis_options, "Algorithm info", "info")



import collections
from contextlib import suppress

import packaging.version
import tifffile.tifffile
from tifffile import TiffFile, TiffPage

if tifffile.tifffile.TiffPage.__module__ != "PartSegImage.tifffile_fixes":
    original_asarray = TiffPage.asarray

    def asarray(self, *args, **kwargs):
        """
        Modified for progress info. call original implementation and send info that page is read by
        call parameter less function `report_func` of parent.
        sample of usage in :py:meth:`ImageRead.read`
        """
        # Because of TiffFrame usage
        res = original_asarray(self, *args, **kwargs)
        self.parent.report_func()
        return res

    def _dummy_report_func(self):  # noqa: ARG001
        """dummy function for report_func"""

    TiffFile.report_func = _dummy_report_func
    tifffile.tifffile.TiffPage.asarray = asarray

    if packaging.version.parse(tifffile.__version__) <= packaging.version.parse("2019.7.26"):  # pragma: no cover
        asbool = tifffile.tifffile.asbool

        def _xml2dict(xml, sanitize=True, prefix=None):
            """Return XML as dict.

            >>> _xml2dict('<?xml version="1.0" ?><root attr="name"><key>1</key></root>')
            {'root': {'key': 1, 'attr': 'name'}}

            """
            from defusedxml import cElementTree as etree  # delayed import

            at = tx = ""
            if prefix:
                at, tx = prefix

            def astype(value):
                # return value as int, float, bool, or str
                if not isinstance(value, str):
                    return value

                for t in (int, float, asbool):
                    with suppress(TypeError, ValueError):
                        return t(value)
                return value

            def etree2dict(t):
                # adapted from https://stackoverflow.com/a/10077069/453463
                key = t.tag
                if sanitize:
                    key = key.rsplit("}", 1)[-1]
                d = {key: {} if t.attrib else None}
                children = list(t)
                if children:
                    dd = collections.defaultdict(list)
                    for dc in map(etree2dict, children):
                        for k, v in dc.items():
                            dd[k].append(astype(v))
                    d = {key: {k: astype(v[0]) if len(v) == 1 else astype(v) for k, v in dd.items()}}
                if t.attrib:
                    d[key].update((at + k, astype(v)) for k, v in t.attrib.items())
                if t.text:
                    text = t.text.strip()
                    if children or t.attrib:
                        if text:
                            d[key][f"{tx}value"] = astype(text)
                    else:
                        d[key] = astype(text)
                return d

            return etree2dict(etree.fromstring(xml))

        tifffile.xml2dict = _xml2dict
        tifffile.tifffile.xml2dict = _xml2dict
        import czifile

        czifile.czifile.xml2dict = _xml2dict



import os

from qtpy.QtGui import QIcon

from PartSegData import icons_dir


class IconSelector:
    def __init__(self):
        self._close_icon = None
        self._edit_icon = None

    @property
    def close_icon(self) -> QIcon:
        if self._close_icon is None:
            self._close_icon = QIcon(os.path.join(icons_dir, "task-reject.png"))
        return self._close_icon

    @property
    def edit_icon(self):
        if self._edit_icon is None:
            self._edit_icon = QIcon(os.path.join(icons_dir, "configure.png"))
        return self._edit_icon



import operator
from enum import Enum
from typing import Optional, Type

import numpy as np
import SimpleITK as sitk
from napari import Viewer
from napari.layers import Image as NapariImage
from napari.layers import Labels, Layer
from napari.utils.notifications import show_info
from pydantic import ValidationError
from qtpy.QtWidgets import QPushButton, QVBoxLayout, QWidget

from PartSeg.plugins import register as register_plugins
from PartSeg.plugins.napari_widgets._settings import get_settings
from PartSeg.plugins.napari_widgets.utils import NapariFormWidget
from PartSegCore.segmentation.border_smoothing import SmoothAlgorithmSelection
from PartSegCore.segmentation.noise_filtering import NoiseFilterSelection
from PartSegCore.segmentation.threshold import DoubleThresholdSelection, ThresholdSelection
from PartSegCore.segmentation.watershed import WatershedSelection
from PartSegCore.utils import BaseModel
from PartSegImage.image import minimal_dtype


class CompareType(Enum):
    lower_threshold = 1
    upper_threshold = 2

    def __str__(self):
        return self.name.replace("_", " ").title()


class FlowType(Enum):
    bright_center = 1
    dark_center = 2

    def __str__(self):
        return self.name.replace("_", " ").title()


class AlgModel(BaseModel):
    layer_name: str = ""

    class Config(BaseModel.Config):
        arbitrary_types_allowed = True

    def run_calculation(self):
        raise NotImplementedError


class ThresholdModel(AlgModel):
    class Config(BaseModel.Config):
        arbitrary_types_allowed = True

    data: NapariImage
    mask: Optional[Labels] = None
    operator: CompareType = CompareType.lower_threshold
    threshold: ThresholdSelection = ThresholdSelection.get_default()

    def run_calculation(self):
        data = np.squeeze(self.data.data)
        mask = np.squeeze(self.mask.data) if self.mask is not None else None
        if mask is not None and mask.shape != data.shape:
            raise ValueError("Mask shape is different than data shape")
        op = operator.gt if self.operator == CompareType.lower_threshold else operator.lt
        res_data = self.threshold.algorithm().calculate_mask(
            data=data, mask=mask, arguments=self.threshold.values, operator=op
        )[0]
        return {
            "data": res_data.reshape(self.data.data.shape),
            "meta": {"scale": self.data.scale, "name": self.layer_name or "Threshold labels"},
            "layer_type": "labels",
        }


class DoubleThresholdModel(ThresholdModel):
    threshold: DoubleThresholdSelection = DoubleThresholdSelection.get_default()


class NoiseFilterModel(AlgModel):
    data: NapariImage
    noise_filtering: NoiseFilterSelection = NoiseFilterSelection.get_default()

    def run_calculation(self):
        data = np.squeeze(self.data.data)
        res_data = self.noise_filtering.algorithm().noise_filter(
            channel=data, spacing=self.data.scale[-3:], arguments=self.noise_filtering.values
        )
        return {
            "data": res_data.reshape(self.data.data.shape),
            "meta": {
                "scale": self.data.scale,
                "contrast_limits": self.data.contrast_limits,
                "name": self.layer_name or "Denoised image",
            },
            "layer_type": "image",
        }


def make_3d(data: np.ndarray):
    if data.ndim > 3:
        data = np.squeeze(data)
    if data.ndim == 2:
        return data.reshape((1, *data.shape))
    if data.ndim != 3:
        raise ValueError(f"Wrong data dimensionality with shape {data.shape}")
    return data


class WatershedModel(AlgModel):
    data: NapariImage
    flow_area: Labels
    core_objects: Labels
    mask: Optional[Labels] = None
    watershed: WatershedSelection = WatershedSelection.get_default()
    side_connection: bool = True
    operator: FlowType = FlowType.bright_center

    def run_calculation(self):
        data = make_3d(self.data.data)
        if self.flow_area is self.core_objects:
            flow_area = make_3d(self.flow_area.data == 1).astype(np.uint8)
            core_objects = sitk.GetArrayFromImage(
                sitk.ConnectedComponent(sitk.GetImageFromArray(make_3d(self.flow_area.data == 2).astype(np.uint8)))
            )
        else:
            flow_area = make_3d(self.flow_area.data)
            core_objects = make_3d(self.core_objects.data)
        components_num = np.amax(core_objects)
        core_objects = core_objects.astype(minimal_dtype(components_num))
        op = operator.gt if self.operator == FlowType.bright_center else operator.lt

        if op(1, 0):
            lower_bound = np.min(data[flow_area > 0])
            upper_bound = np.max(data[flow_area > 0])
        else:
            lower_bound = np.max(data[flow_area > 0])
            upper_bound = np.min(data[flow_area > 0])
        data = self.watershed.algorithm().sprawl(
            data=data,
            sprawl_area=flow_area,
            core_objects=core_objects,
            components_num=components_num,
            arguments=self.watershed.values,
            spacing=self.flow_area.scale[-3:],
            side_connection=self.side_connection,
            operator=op,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
        )
        return {
            "data": data.reshape(self.flow_area.data.shape),
            "meta": {
                "scale": self.flow_area.scale,
                "name": self.layer_name or "Watershed",
            },
            "layer_type": "labels",
        }


class BorderSmoothingModel(AlgModel):
    data: Labels
    border_smoothing: SmoothAlgorithmSelection = SmoothAlgorithmSelection.get_default()
    only_side_connection: bool = True

    def run_calculation(self):
        data = self.data.data
        res_data = self.border_smoothing.algorithm().smooth(
            segmentation=np.squeeze(data), arguments=self.border_smoothing.values
        )
        return {
            "data": res_data.reshape(self.data.data.shape),
            "meta": {
                "scale": self.data.scale,
                "name": self.layer_name or "Border smoothed",
            },
            "layer_type": "labels",
        }


class ConnectedComponentsModel(AlgModel):
    data: Labels
    side_connection: bool = True
    minimum_size: int = 20

    def run_calculation(self):
        data = np.squeeze(self.data.data)
        res_data = sitk.GetArrayFromImage(
            sitk.RelabelComponent(
                sitk.ConnectedComponent(sitk.GetImageFromArray(data), self.side_connection),
                self.minimum_size,
            )
        )
        return {
            "data": res_data.reshape(self.data.data.shape),
            "meta": {
                "scale": self.data.scale,
                "name": self.layer_name or "Connected components",
            },
            "layer_type": "labels",
        }


class SplitCoreObjectsModel(AlgModel):
    data: Labels

    def run_calculation(self):
        res_data = (self.data.data >= 2).astype(np.uint8)
        return {
            "data": res_data,
            "meta": {
                "scale": self.data.scale,
                "name": self.layer_name or "Core objects",
            },
            "layer_type": "labels",
        }


class AlgorithmWidgetBase(QWidget):
    __data_model__: Type[BaseModel]

    def __init__(self, napari_viewer: Viewer, parent=None):
        super().__init__(parent)
        self.settings = get_settings()
        self.napari_viewer = napari_viewer

        self.run_button = QPushButton("Run")
        self.run_button.clicked.connect(self.run_operation)

        self.form = NapariFormWidget(self.__data_model__)
        self.form.set_values(self.settings.get(f"widgets.{self.__class__.__name__}", {}))

        layout = QVBoxLayout()
        layout.addWidget(self.form)
        layout.addWidget(self.run_button)
        self.setLayout(layout)
        register_plugins()

    def reset_choices(self, event=None):
        self.form.reset_choices(event)

    def run_operation(self):
        try:
            res = self.form.get_values().run_calculation()
        except ValidationError:
            show_info("It looks like not all layers are selected")
            return
        layer = Layer.create(**res)
        if (
            layer.name in self.napari_viewer.layers
            and layer.__class__ is self.napari_viewer.layers[layer.name].__class__
            and np.array_equal(layer.scale, self.napari_viewer.layers[layer.name].scale)
        ):
            self.napari_viewer.layers[layer.name].data = res["data"]
        else:
            self.napari_viewer.add_layer(layer)
        self.settings.set(f"widgets.{self.__class__.__name__}", dict(self.form.get_values()))
        self.settings.dump()


class Threshold(AlgorithmWidgetBase):
    __data_model__ = ThresholdModel


class DoubleThreshold(AlgorithmWidgetBase):
    __data_model__ = DoubleThresholdModel


class NoiseFilter(AlgorithmWidgetBase):
    __data_model__ = NoiseFilterModel


class BorderSmooth(AlgorithmWidgetBase):
    __data_model__ = BorderSmoothingModel


class Watershed(AlgorithmWidgetBase):
    __data_model__ = WatershedModel


class ConnectedComponents(AlgorithmWidgetBase):
    __data_model__ = ConnectedComponentsModel


class SplitCoreObjects(AlgorithmWidgetBase):
    __data_model__ = SplitCoreObjectsModel



import json
import os
import sys
import urllib.error
import urllib.request
from contextlib import suppress
from datetime import date

import packaging.version
import sentry_sdk
from qtpy.QtCore import QThread
from qtpy.QtWidgets import QMessageBox
from superqt import ensure_main_thread

from PartSeg import __version__, state_store

IGNORE_DAYS = 21
IGNORE_FILE = "ignore.txt"


class CheckVersionThread(QThread):
    """Thread to check if there is new PartSeg release. Checks base on newest version available on pypi_

    .. _PYPI: https://pypi.org/project/PartSeg/
    """

    def __init__(
        self, package_name="PartSeg", default_url="https://4dnucleome.cent.uw.edu.pl/PartSeg/", base_version=__version__
    ):
        super().__init__()
        self.release = base_version
        self.base_release = base_version
        self.package_name = package_name
        self.url = default_url
        self.finished.connect(self.show_version_info)

    def run(self):
        """This function perform check"""

        # noinspection PyBroadException
        if not state_store.check_for_updates:
            return
        try:
            if os.path.exists(os.path.join(state_store.save_folder, IGNORE_FILE)):
                with open(os.path.join(state_store.save_folder, IGNORE_FILE), encoding="utf-8") as f_p, suppress(
                    ValueError
                ):
                    old_date = date.fromisoformat(f_p.read())
                    if (date.today() - old_date).days < IGNORE_DAYS:
                        return
                os.remove(os.path.join(state_store.save_folder, IGNORE_FILE))

            with urllib.request.urlopen(f"https://pypi.org/pypi/{self.package_name}/json") as r:  # nosec  # noqa: S310
                data = json.load(r)
            self.release = data["info"]["version"]
            self.url = data["info"]["home_page"]
        except (KeyError, urllib.error.URLError):  # pragma: no cover
            pass
        except Exception as e:  # pylint: disable=broad-except
            with sentry_sdk.push_scope() as scope:
                scope.set_tag("auto_report", "true")
                scope.set_tag("check_version", "true")
                sentry_sdk.capture_exception(e)

    @ensure_main_thread
    def show_version_info(self):
        my_version = packaging.version.parse(self.base_release)
        remote_version = packaging.version.parse(self.release)
        if remote_version > my_version:
            if getattr(sys, "frozen", False):
                message = QMessageBox(
                    QMessageBox.Icon.Information,
                    "New release",
                    f"You use outdated version of PartSeg. "
                    f"Your version is {my_version} and current is {remote_version}. "
                    f"You can download next release form {self.url}",
                    QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Ignore,
                )
            else:
                message = QMessageBox(
                    QMessageBox.Icon.Information,
                    "New release",
                    f"You use outdated version of PartSeg. "
                    f"Your version is {my_version} and current is {remote_version}. "
                    "You can update it from pypi (pip install -U PartSeg)",
                    QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Ignore,
                )

            if message.exec_() == QMessageBox.StandardButton.Ignore:
                os.makedirs(state_store.save_folder, exist_ok=True)
                with open(os.path.join(state_store.save_folder, IGNORE_FILE), "w", encoding="utf-8") as f_p:
                    f_p.write(date.today().isoformat())



from PartSegCore.mask.io_functions import MaskProjectTuple
from PartSegCore.mask_create import MaskProperty
from PartSegCore.project_info import HistoryElement


def create_history_element_from_segmentation_tuple(project_info: MaskProjectTuple, mask_property: MaskProperty):
    return HistoryElement.create(
        roi_info=project_info.roi_info,
        mask=project_info.mask,
        roi_extraction_parameters={
            "selected": project_info.selected_components,
            "parameters": project_info.roi_extraction_parameters,
        },
        mask_property=mask_property,
    )



from PartSegCore.napari_plugins import (
    load_image,
    load_mask_project,
    load_masked_image,
    load_roi_project,
    save_mask_roi,
    save_tiff_layer,
)

__all__ = [
    "load_image",
    "load_mask_project",
    "load_masked_image",
    "load_roi_project",
    "save_mask_roi",
    "save_tiff_layer",
]



import itertools
import os
from pathlib import Path

import pytest

from PartSegCore.analysis import ProjectTuple
from PartSegCore.analysis.load_functions import LoadImageMask, LoadMaskSegmentation, LoadProject, LoadStackImage
from PartSegCore.analysis.save_functions import save_dict as analysis_save_dict
from PartSegCore.mask.io_functions import (
    LoadROIImage,
    LoadROIParameters,
    MaskProjectTuple,
    save_components_dict,
    save_parameters_dict,
    save_segmentation_dict,
)
from PartSegCore.mask.io_functions import LoadStackImage as MaskLoadStackImage


class TestPartSettingsIO:
    @pytest.mark.parametrize("method", analysis_save_dict.values())
    def test_save(self, method, part_settings_with_project, tmp_path):
        pi = part_settings_with_project.get_project_info()
        method.save(tmp_path / "data", pi, method.get_default_values())

    def test_load_segmentation(self, part_settings, data_test_dir):
        path = os.path.join(data_test_dir, "stack1_component1_1.tgz")
        pi = LoadProject.load([path])
        part_settings.set_project_info(pi)

    @pytest.mark.parametrize(
        "file_name",
        [
            "Image0003_01.oif",
            "test_czi.czi",
            "test_lsm.lsm",
            "test_lsm.tif",
            "test_lsm2.tif",
            "test_nucleus.tif",
            "test_nucleus_mask.tif",
        ],
    )
    def test_load_images(self, file_name, part_settings, data_test_dir):
        path = os.path.join(data_test_dir, file_name)
        pi = LoadStackImage.load([path])
        assert isinstance(pi, ProjectTuple)
        part_settings.set_project_info(pi)

    def test_load_images_with_mask(self, part_settings, data_test_dir):
        base_path = Path(os.path.join(data_test_dir, "stack1_components"))

        pi = LoadImageMask.load([base_path / "stack1_component1.tif", base_path / "stack1_component1_mask.tif"])
        assert isinstance(pi, ProjectTuple)
        part_settings.set_project_info(pi)

    def test_load_mask_project(self, part_settings, data_test_dir):
        path = os.path.join(data_test_dir, "test_nucleus.seg")
        pi_l = LoadMaskSegmentation.load([path])
        assert isinstance(pi_l, list)
        for pi in pi_l:
            part_settings.set_project_info(pi)


class TestStackSettingsIO:
    @pytest.mark.parametrize(
        "method",
        itertools.chain(save_segmentation_dict.values(), save_parameters_dict.values(), save_components_dict.values()),
    )
    def test_save(self, method, stack_segmentation1, stack_settings, tmp_path):
        stack_settings.set_project_info(stack_segmentation1)
        pi = stack_settings.get_project_info()
        method.save(tmp_path / "data", pi, method.get_default_values())

    @pytest.mark.parametrize(
        "file_name",
        [
            "Image0003_01.oif",
            "test_czi.czi",
            "test_lsm.lsm",
            "test_lsm.tif",
            "test_lsm2.tif",
            "test_nucleus.tif",
            "test_nucleus_mask.tif",
        ],
    )
    def test_load_images(self, file_name, stack_settings, data_test_dir):
        path = os.path.join(data_test_dir, file_name)
        pi = MaskLoadStackImage.load([path])
        assert isinstance(pi, MaskProjectTuple)
        stack_settings.set_project_info(pi)

    def test_load_segmentation_image(self, data_test_dir, stack_settings):
        pi = LoadROIImage.load([os.path.join(data_test_dir, "test_nucleus.seg")])
        assert isinstance(pi, MaskProjectTuple)
        stack_settings.set_project_info(pi)
        assert pi.image is not None

    def test_load_parameters(self, data_test_dir):
        pi = LoadROIParameters.load([os.path.join(data_test_dir, "test_nucleus.seg")])
        assert isinstance(pi, MaskProjectTuple)
        assert len(pi.roi_extraction_parameters) == 4
        assert pi.roi_extraction_parameters[1].algorithm == "Threshold"



import functools

from PartSegCore.mask.io_functions import LoadROI
from PartSegCore.napari_plugins.loader import partseg_loader


def napari_get_reader(path: str):
    return next(
        (
            functools.partial(partseg_loader, LoadROI)
            for extension in LoadROI.get_extensions()
            if path.endswith(extension)
        ),
        None,
    )
